<?php
/**
* @package msta emailontopic Extension
*/

namespace msta\emailontopic\acp;

class main_module {
  public function main () {
    global $db, $user, $request, $template, $config, $phpbb_root_path, $phpbb_admin_path, $phpEx, $phpbb_container;

    $this->db = $db;
    $this->user = $user;
    $this->template = $template;
    $this->config = $config;
    $this->request = $request;
    $this->phpbb_root_path = $phpbb_root_path;
    $this->phpbb_admin_path = $phpbb_admin_path;
    $this->phpex = $phpEx;
    $this->helper = $phpbb_container->get('controller.helper');
    $this->ext_root_path = $phpbb_root_path.'ext/msta/emailontopic/';
    $this->user->add_lang('acp/common');
    $this->page_title = $this->user->lang('ACP_EMAILONTOPIC');
    $this->table_prefix = substr(USERS_TABLE, 0, -5);

    $this->template->assign_vars(array('FORM_ACTION' => $this->u_action));

    switch ($this->request->variable('mode', 'group_settings')) {
      case 'group_settings':
        $this->mail_group_settings();
        break;
      case 'telegram_settings':
        $this->telegram_settings();
        break;
    }
  }

  private function urlToWebhook ($pwd) {
    $path = explode('?', $this->helper->route('msta_emailontopic_controller'));
    if (count($path) > 1) {
      $query = explode('&', $path[1]);
      for ($i = 0; $i < count($query); $i++)
        if (substr($query[$i], 0, 4) == 'sid=') {
          array_splice($query, $i, 1);
          break;
        }
    } else $query = array();
    $query[] = 'action=telegraminput';
    $query[] = 'pwd='.rawurlencode($pwd);
    $path = $path[0].'?'.join('&', $query);
    return (($this->request->is_secure())?'https://':'http://').$this->request->server('HTTP_HOST').$path;
  }

  private function generatePassword ($length) {
    $result = '';
    for ($i = 0; $i < $length; $i++) {
      $c = mt_rand(0, 61);
      if ($c < 10) $result .= chr($c+48);
      else if ($c < 36) $result .= chr($c-10+65);
      else $result .= chr($c-36+97);
    }
    return $result;
  }

  private function telegramOp ($opQuery) {
    $result = null;
    $context = stream_context_create(array('http' => array('ignore_errors' => true)));
    $response = @file_get_contents('https://api.telegram.org/bot'.$this->config['telegram_bot_key'].'/'.$opQuery, false, $context);
    if ($response) {
      try {
        $result = json_decode($response, true);
      } catch (Exception $e) { $result = null; }
    }
    return $result;
  }

  private function sqlInsertStr ($obj) {
    $keyStr = '';
    $valStr = '';
    foreach ($obj as $k => $v) {
      if ($keyStr != '') {
        $keyStr .= ', ';
        $valStr .= ', ';
      }
      $keyStr .= $k;
      if (gettype($v) == 'integer') $valStr .= $v;
      else if (gettype($v) == 'string') $valStr .= "'".$this->db->sql_escape($v)."'";
      else die('Unsupported variable type for SQL!');
    }
    return '('.$keyStr.') VALUES ('.$valStr.')';
  }

  private function createTelegramUser () {
    $info = array(
      'user_type' => 2,
      'group_id' => 6,
      'user_regdate' => date('U'),
      'username' => 'telegram',
      'username_clean' => 'telegram',
      'user_permissions' => '',
      'user_sig' => '',
    );
    $lastId = 0;
    if ($this->db->sql_query('INSERT INTO '.$this->table_prefix.'users '.$this->sqlInsertStr($info)))
      $lastId = intval($this->db->sql_nextid());
    if ($lastId == 0) die('Cannot create telegram user!');
    $this->config->set('telegram_user_id', $lastId);
    return $lastId;
  }

  private function telegram_settings () {
    if (intval($this->config['telegram_user_id']) == 0) $this->createTelegramUser();
    $this->tpl_name = 'acp_telegram_settings';
    $msg = '';
    $botStatus = null;

    if ($this->request->variable('but_activate', 'no') != 'no') {
      $botKey = $this->request->variable('telegram_bot_key', '');
      $this->config->set('telegram_bot_key', $botKey);
      if (strpos($botKey, 'bot') === 0) $botKey = substr($botKey, 3);
      $pwd = $this->config['telegram_webhook_pass'];
      if ($pwd) $msg = 'Webhook already set!';
      else if ($botKey == '') $msg = 'Bot key not specified!';
      else {
        $botStatus = $this->telegramOp('getMe');
        if ($botStatus && $botStatus['ok']) {
          $pwd = $this->generatePassword(20);
          $hookSetup = $this->telegramOp('setWebhook?url='.rawurlencode($this->urlToWebhook($pwd)));
          if ($hookSetup && $hookSetup['ok']) {
            $msg = 'Webhook OK.';
            $this->config->set('telegram_webhook_pass', $pwd);
          } else $msg = 'Webhook setup failed!'.(($hookSetup)?'  ['.htmlspecialchars($hookSetup['description']).']':'');
        } else $msg = 'Bot not connected!';
      }
    }

    if ($this->request->variable('but_deactivate', 'no') != 'no') {
      $pwd = $this->config['telegram_webhook_pass'];
      if (!$pwd) $msg = 'Webhook already deactivated!';
      else {
        $this->telegramOp('deleteWebhook?drop_pending_updates=1');
        $this->config->set('telegram_webhook_pass', '');
      }
    }

    $status = '';
    if (!$botStatus) $botStatus = $this->telegramOp('getMe');
    $webhookStatus = $this->telegramOp('getWebhookInfo');
    if ($botStatus && $botStatus['result']) {
      $botStatus['result']['ok'] = $botStatus['ok'];
      $botStatus = $botStatus['result'];
      foreach ($botStatus as $k => $v) {
        if ($k == 'id') $this->config->set('telegram_bot_id', intval($v));
        $status .= '<tr><td>Bot</td><td>'.htmlspecialchars($k).'</td><td>'.htmlspecialchars(strval($v)).'</td></tr>'."\n";
      }
    }
    if ($webhookStatus && $webhookStatus['result']) {
      $webhookStatus['result']['ok'] = $webhookStatus['ok'];
      $webhookStatus = $webhookStatus['result'];
      foreach ($webhookStatus as $k => $v)
        $status .= '<tr><td>Webhook</td><td>'.htmlspecialchars($k).'</td><td>'.htmlspecialchars(strval($v)).'</td></tr>'."\n";
    }
    $this->template->assign_vars(array(
      'telegram_bot_key' => $this->config['telegram_bot_key'],
      'telegram_bot_active' => ($this->config['telegram_webhook_pass'])?true:false,
      'info_message' => $msg,
      'telegram_bot_status' => $status,
    ));
  }

  private function mail_group_settings () {
    $this->tpl_name = 'acp_email_settings';

    // List groups and create json information
    $groups_response = $this->db->sql_query('SELECT group_id, group_name FROM '.$this->table_prefix.'groups ORDER BY group_id');
    $json = '[';
    $groupIds = array();
    while ($row = $this->db->sql_fetchrow($groups_response)) {
      if ($json != '[') $json .= ', ';
      $json .= '{"id":"'.$row['group_id'].'", "gName":"'.htmlspecialchars($row['group_name']).'"}';
      $groupIds[] = intval($row['group_id']);
    }
    $json .= ']';

    if ($this->request->variable('but_apply', 'no') != 'no') { // if this is set, the user has submitted the form
      $var_names = $this->request->variable_names(\phpbb\request\request_interface::POST);
      foreach ($var_names as $vn) {
        $spl = explode('_', $vn);
        if (count($spl) != 3 || $spl[0] != 'row' || $spl[2] != 'sender') continue; // regard each sender id
        $row = array(
          'sender' => intval($this->request->variable('row_'.$spl[1].'_sender', 0)),
          'receiver' => intval($this->request->variable('row_'.$spl[1].'_receiver', 0)),
          'nrecvM' => intval($this->request->variable('row_'.$spl[1].'_nrecv_m', 0)),
          'nrecvL' => intval($this->request->variable('row_'.$spl[1].'_nrecv_l', 0)),
          'subscriptionMgr' => ($this->request->variable('row_'.$spl[1].'_subsmgr_m', 'no') != 'no')?2:(($this->request->variable('row_'.$spl[1].'_subsmgr_l', 'no') != 'no')?1:0),
        );
        if ($row['nrecvM'] > 255) $row['nrecvM'] = 255;
        if ($row['nrecvL'] > 255) $row['nrecvL'] = 255;
        if ($row['nrecvM'] < 0) $row['nrecvM'] = 0;
        if ($row['nrecvL'] < 0) $row['nrecvL'] = 0;
        if ((array_search($row['sender'], $groupIds) === false) || (array_search($row['receiver'], $groupIds) === false)) { // entries for non-existing groups can only be deleted
          $row['nrecvM'] = 0;
          $row['nrecvL'] = 0;
        }
        if ($row['sender'] <= 0 || $row['receiver'] <= 0) continue; // invalid key
        if ($row['nrecvM'] == 0 && $row['nrecvL'] == 0) { // delete this entry
          $this->db->sql_query('DELETE FROM '.$this->table_prefix.'group_mail_privileges WHERE from_group_id='.$row['sender'].' AND to_group_id='.$row['receiver']);
        } else { // add or update this entry
          $groups_response = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'group_mail_privileges WHERE from_group_id='.$row['sender'].' AND to_group_id='.$row['receiver']);
          $exists = false;
          while ($res = $this->db->sql_fetchrow($groups_response)) $exists = true;
          if ($exists) $this->db->sql_query('UPDATE '.$this->table_prefix.'group_mail_privileges SET user_recipient_count = '.$row['nrecvM'].', leader_recipient_count = '.$row['nrecvL'].', subscription_management = '.$row['subscriptionMgr'].' WHERE from_group_id='.$row['sender'].' AND to_group_id='.$row['receiver']);
          else $this->db->sql_query('INSERT INTO '.$this->table_prefix.'group_mail_privileges (from_group_id, to_group_id, user_recipient_count, leader_recipient_count, subscription_management) VALUES ('.$row['sender'].', '.$row['receiver'].', '.$row['nrecvM'].', '.$row['nrecvL'].', '.$row['subscriptionMgr'].')');
        }
      }
    }

    // Create permissions table
    $groups_response = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'group_mail_privileges');
    $row_nr = 0;
    $this->template->assign_block_vars('group', array( // first row is template for new rows
      'row' => $row_nr,
      'sender' => 0,
      'receiver' => 0,
      'memberNr' => 3,
      'leaderNr' => 0,
      'memberSubsMgrChecked' => '',
      'leaderSubsMgrChecked' => '',
    ));
    while ($row = $this->db->sql_fetchrow($groups_response)) {
      $row_nr++;
      $this->template->assign_block_vars('group', array(
        'row' => $row_nr,
        'sender' => $row['from_group_id'],
        'receiver' => $row['to_group_id'],
        'memberNr' => intval($row['user_recipient_count']),
        'leaderNr' => intval($row['leader_recipient_count']),
        'memberSubsMgrChecked' => ($row['subscription_management'] == 2)?'checked':'',
        'leaderSubsMgrChecked' => ($row['subscription_management'] >= 1)?'checked':'',
      ));
    }
    // store group information for js
    $this->template->assign_vars(array('GROUP_LIST' => $json));
  }
}
