<?php

namespace msta\emailontopic\event;

use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class main_listener implements EventSubscriberInterface {
  protected $phpbb_container;
  protected $helper;
  protected $template;
  protected $user;
  protected $db;
  protected $request;
  protected $auth;
  protected $config;
  protected $phpbb_dispatcher;
  protected $language;
  protected $phpbb_root_path;
  private $table_prefix;
  private $permissionCheck = null;
  private $bridgeManager = null;
  private $postingParams = false;

  public function __construct ($phpbb_container, \phpbb\controller\helper $helper, \phpbb\template\template $template, \phpbb\user $user, \phpbb\db\driver\driver_interface $db, \phpbb\request\request $request, \phpbb\auth\auth $auth, \phpbb\config\config $config, \phpbb\event\dispatcher_interface $phpbb_dispatcher, \phpbb\language\language $language, $phpbb_root_path) {
    $this->phpbb_container = $phpbb_container;
    $this->helper = $helper;
    $this->template = $template;
    $this->user = $user;
    $this->db = $db;
    $this->request = $request;
    $this->auth = $auth;
    $this->config = $config;
    $this->phpbb_dispatcher = $phpbb_dispatcher;
    $this->language = $language;
    $this->phpbb_root_path = $phpbb_root_path;
    $this->table_prefix = substr(USERS_TABLE, 0, -5);
  }

  static public function getSubscribedEvents () {
    return array(
      'core.permissions' => 'register_permissions',
      'core.delete_posts_after' => 'delete_posts',
      'core.submit_post_end' => 'submit_post_end',
      'core.modify_posting_parameters' => 'modify_posting_parameters',
      'core.notification_manager_add_notifications' => 'notification_manager_add_notifications',
      'core.modify_email_headers' => 'modify_mail_headers',
      'core.page_header' => 'page_header',
      'core.ucp_prefs_personal_data' => 'show_ucp_settings',
      'core.ucp_prefs_personal_update_data' => 'post_ucp_settings',
      'core.user_add_modify_notifications_data' => 'new_user_notifications',
      'core.viewtopic_get_post_data' => 'viewtopic',
      'core.user_add_after' => 'new_user_mod',
      'core.acp_manage_forums_display_form' => 'forum_security_settings_show',
      'core.acp_manage_forums_update_data_before' => 'forum_security_settings_set',
      'core.acp_board_config_edit_add' => 'general_acp_settings',
      'msta.emailontopic.input_bridge_post_end' => 'input_bridge_post_end',
      //'core.modify_notification_message' => 'display_mail_send',  // this can be used for debugging to show the mail after submit in browser
    );
  }

  public function pgp_enabled_gui ($value, $key) {
    return
       '<label><input type="radio" name="config[emailontopic_gnupg_enabled]" value="2" '.(($value == 2)?'checked="checked"':'').' class="radio" '.((class_exists('\gnupg'))?'':'disabled="disabled"').'> GNUPG</label>'
      .'<label><input type="radio" name="config[emailontopic_gnupg_enabled]" value="1" '.(($value == 1)?'checked="checked"':'').' class="radio"> OpenPGP-PHP</label>'
      .'<label><input type="radio" name="config[emailontopic_gnupg_enabled]" value="0" '.(($value == 0)?'checked="checked"':'').' class="radio" id="emailontopic_gnupg_enabled"> '.$this->user->lang['DISABLED'].'</label>'
      .'<br>'
      .((class_exists('\gnupg'))?'<span style="color:#0A0;">'.$this->user->lang['EMAILONTOPIC_PGP_AVAILLABLE']:'<span style="color:#A00;">'.$this->user->lang['EMAILONTOPIC_PGP_NOT_AVAILLABLE']).'</span>';
  }

  public function general_acp_settings ($event) {
    if ($event['mode'] == 'load') {
      $this->user->add_lang_ext('msta/emailontopic', 'ucp_emailontopic');
      $v = $event['display_vars'];
      $nv = array();
      foreach ($v['vars'] as $key => $val) {
        $nv[$key] = $val;
        if ($key == 'session_length') {
          $nv['emailontopic_keep_connection'] = array('lang' => 'EMAILONTOPIC_KEEP_CONNECTION', 'validate' => 'bool', 'type' => 'radio:yes_no', 'explain' => 1);
        }
      }
      $v['vars'] = $nv;
      $event['display_vars'] = $v;
    } else if ($event['mode'] == 'email') {
      if ($event['submit'] == 1) {
        $file = $this->request->file('pgpFile');
        if ($file && $file['tmp_name']) {
          $mailMessenger = \msta\emailontopic\includes\bridges\MailMessenger::getInstance($this->phpbb_container, $this->db, $this->config, $this->phpbb_dispatcher, $this->phpbb_root_path);
          $mailMessenger->importPgpKey($file['tmp_name'], 0);
        }
        if ($this->request->variable('pgpDelete', 'no') != 'no') {
          $this->config->set('emailontopic_forum_pgp_key', '');
        }
      }
      $this->user->add_lang_ext('msta/emailontopic', 'ucp_emailontopic');
      $v = $event['display_vars'];
      $nv = array();
      foreach ($v['vars'] as $key => $val) {
        if ($key == 'legend2') {
          $nv['emailontopic_gnupg_enabled'] = array('lang' => 'EMAILONTOPIC_PGP_ENABLED', 'validate' => 'int:0:2', 'type' => 'custom', 'function' => array($this, 'pgp_enabled_gui'), 'explain' => 1);
          $nv['emailontopic_forum_pgp_key'] = array('lang' => 'EMAILONTOPIC_PGP_BOARD', 'validate' => 'string', 'type' => 'text:25:255', 'explain' => 1);
        }
        $nv[$key] = $val;
      }
      $v['vars'] = $nv;
      $event['display_vars'] = $v;
    }
  }

  public function forum_security_settings_set ($event) {
    $fd = $event['forum_data_sql'];
    $fd['bridging_security'] = (intval($this->request->variable('forum_security_bridging', 0)) & 96) | (($this->request->variable('forum_security_bridging_subject', 'no') == 'no')?0:1);
    $event['forum_data_sql'] = $fd;
  }

  public function forum_security_settings_show ($event) {
    $this->user->add_lang_ext('msta/emailontopic', 'info_acp_forumsettings');
    $this->template->assign_vars([
      'forum_security_bridging' => intval($event['forum_data']['bridging_security']) & 96,
      'forum_security_bridging_subject' => intval($event['forum_data']['bridging_security']) & 1,
    ]);
  }

  public function register_permissions ($event) {
    $event['categories'] = array_merge($event['categories'], [
      'bridging' => 'FORUM_RULES_BRIDGING_TAB',
    ]);
    $event['permissions'] = array_merge($event['permissions'], [
      'f_configtelegram' => ['lang' => 'ACL_F_CONFIGTELEGRAM', 'cat' => 'bridging'],
    ]);
  }

  public function new_user_mod ($event) {
    $this->db->sql_query('UPDATE '.USERS_TABLE.' SET user_notify = 1 WHERE user_id = '.intval($event['user_id']));
  }

  public function display_mail_send ($event) {
    die($event['message']);
  }

  public function viewtopic ($event) {
    $tid = intval($event['topic_data']['topic_id']);
    $this->user->add_lang_ext('msta/emailontopic', 'topic_end');
    // variables for telegram info box
    $telegramLinksResponse = $this->db->sql_query('SELECT chat_name FROM '.$this->table_prefix.'telegram_links WHERE forum_link_id = '.$tid);
    $telegramLinks = array();
    while ($row = $this->db->sql_fetchrow($telegramLinksResponse))
      $telegramLinks[] = htmlspecialchars($row['chat_name']);
    $this->template->assign_vars(['TELEGRAM_LINKS' => join(', ', $telegramLinks)]);
    // variables for subscription buttons
    if (intval($this->user->data['user_id']) != 0) {
      $existingRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT enduring FROM '.$this->table_prefix."topics_watch WHERE topic_id = ".$tid." AND user_id = ".intval($this->user->data['user_id'])));
      $subs = 0;
      if ($existingRow)
        if ($existingRow['enduring'] == 1) $subs = 2;
        else $subs = 1;
      $this->template->assign_vars([
        'SUBSCRIPTION_STATE' => $subs,
        'SUBSCRIPTION_TOPIC_ID' => $tid,
      ]);
    }
  }

  public function new_user_notifications ($event) { // activate email notification on new pm for new users
    for ($i = 0; $i < count($event['notifications_data']); $i++)
      if ($event['notifications_data'][$i]['item_type'] == 'notification.type.pm') return;
    $not = array_merge($event['notifications_data'], array(array(
      'item_type' => 'notification.type.pm',
      'method' => 'notification.method.email',
    )));
    $event['notifications_data'] = $not;
  }

  public function notification_manager_add_notifications ($event) { // blocks default board notifications from being sent
    if ($event['notification_type_name'] == 'notification.type.topic' || $event['notification_type_name'] == 'notification.type.post') {
      $notify_users = $event['notify_users'];
      foreach ($notify_users as $k => $v)
        $notify_users[$k] = array();
      $event['notify_users'] = $notify_users;
    }
  }

  public function modify_mail_headers ($event) { // removes the default plain-text-headers to allow the html type headers.
   /* if (!\msta\emailontopic\includes\MailMessenger::$removeHeaders) return; // do this only for mails generated by this extension
    $headers = $event['headers'];
    for ($i = count($headers)-1; $i >= 0; $i--)
      if (strpos($headers[$i], 'Content-Type:') === 0 || strpos($headers[$i], 'X-Priority:') === 0 || strpos($headers[$i], 'X-MSMail-Priority:') === 0)
        array_splice($headers, $i, 1);
    $event['headers'] = $headers;*/
  }

  public function show_ucp_settings ($event) {
    $this->user->add_lang_ext('msta/emailontopic', 'ucp_emailontopic');
    $sqlResponse = $this->db->sql_fetchrow($this->db->sql_query("SELECT pf_pgp_fingerprint FROM ".$this->table_prefix."profile_fields_data WHERE user_id = ".intval($this->user->data['user_id'])));
    $pgpFingerprint = '';
    if ($sqlResponse) $pgpFingerprint = $sqlResponse['pf_pgp_fingerprint'];
    $this->template->assign_vars([
      'EmailOverride' => intval($this->user->data['mail_visible_override']),
      'pgpFingerprint' => $pgpFingerprint,
    ]);
  }

  public function post_ucp_settings ($event) {
    $override = $this->request->variable('EmailOverride', 0);
    if ($this->request->variable('viewemail', 0) == 0) $override = 1;
    $this->db->sql_query('UPDATE '.$this->table_prefix.'users SET mail_visible_override = '.intval($override).' WHERE user_id = '.intval($this->user->data['user_id']));
    $file = $this->request->file('pgpFile');
    if ($file && $file['tmp_name'] && $this->user->data['user_id'] > 0) {
      $mailMessenger = \msta\emailontopic\includes\bridges\MailMessenger::getInstance($this->phpbb_container, $this->db, $this->config, $this->phpbb_dispatcher, $this->phpbb_root_path);
      $mailMessenger->importPgpKey($file['tmp_name'], $this->user->data['user_id']); // Ensure, user_id is not null here, as this would enable users to replace the forums private key.
    }
    if ($this->request->variable('pgpDelete', 'no') != 'no') {
      $this->db->sql_query("UPDATE ".$this->table_prefix."profile_fields_data SET pf_pgp_fingerprint = '' WHERE user_id = ".intval($this->user->data['user_id']));
    }
  }

  public function page_header ($event) {
    if ($this->postingParams) $this->user->add_lang_ext('msta/emailontopic', 'edit_post_page');
    if (isset($_COOKIE['EMAILONTOPIC'])) {
      $this->user->add_lang_ext('msta/emailontopic', 'report_cookie');
      $this->template->assign_vars(['EMAILONTOPIC_HINT_AFTER_SENT' => 'yes']);
    }
    if (intval($this->config['emailontopic_keep_connection']) == 1) {
      $this->user->add_lang_ext('msta/emailontopic', 'report_onlinestatus');
      $this->template->assign_vars([
        'EMAILONTOPIC_KEEP_CONNECTION' => 'yes',
        'EMAILONTOPIC_UID' => intval($this->user->data['user_id']),
      ]);
    }
  }

  private function getTopicId () {
    $topicId = intval($this->request->variable('t', '', true));
    if ($topicId != 0) return $topicId;
    $postId = intval($this->request->variable('p', '', true));
    $sqlResponse = $this->db->sql_fetchrow($this->db->sql_query('SELECT topic_id FROM '.$this->table_prefix.'posts WHERE post_id = '.$postId));
    if ($sqlResponse) $topicId = intval($sqlResponse['topic_id']);
    return $topicId;
  }

  public function modify_posting_parameters ($event) {
    $this->postingParams = true;
    $this->request->enable_super_globals();
    $topicId = $this->getTopicId();
    // get telegram channels
    $telegramChannels = '';
    $sqlResponse = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_links WHERE (forum_link_id = '.$topicId.' AND link_type < 4 AND link_type > 0) OR (link_type = 0 AND active = 1) ORDER BY join_date');
    while ($row = $this->db->sql_fetchrow($sqlResponse)) {
      $linkType = intval($row['link_type']);
      $isSetForThisTopic = $row['forum_link_id'] == $topicId && $row['link_type'] < 4 && $row['link_type'] > 0;
      $selected = $isSetForThisTopic;
      if ($this->request->variable('preview', 'no') != 'no') { // get selection and link_type from post form, if preview submit
        $selected = false;
        $postTelegramCh = $_POST['telegramChannels'];
        if ($postTelegramCh) foreach ($postTelegramCh as $val) {
          $spl = explode('|', $val);
          if (intval($spl[0]) == intval($row['telegram_chat_id'])) {
            $selected = true;
            if (count($spl) > 1) $linkType = intval($spl[1]);
            break;
          }
        }
      }
      $telegramChannels .= '<option'
       .(($selected)?' selected':'')
       .' currentlySet="'.(($isSetForThisTopic)?'yes':'no').'"'
       .' value="'.intval($row['telegram_chat_id']).'|'.$linkType.'"'
       .' joinDate="'.intval($row['join_date']).'"'
       .' active="'.intval($row['active']).'"'
       .'>'.$row['chat_name'].'</option>';
    }
    // get email list
    $mailEnabled = '';
    $list = '';
    if ($this->request->variable('email_members', '', true) == 'on') $mailEnabled = ' checked';
    if ($this->request->variable('mailSubscriptions', '', true)) $list = $this->request->variable('mailSubscriptions', '', true);
    else {
      $subscribedusers = $this->db->sql_query('SELECT u.username, u.user_id FROM '.$this->table_prefix.'topics_watch AS w INNER JOIN '.$this->table_prefix.'users AS u ON u.user_id = w.user_id WHERE w.topic_id = '.$topicId.' AND enduring = 1');
      $mailtolist = array();
      while ($row = $this->db->sql_fetchrow($subscribedusers)) {
        $mailtolist['U'.intval($row['user_id'])] = [
          'name' => $row['username'],
          'isNew' => false,
          'isGroup' => false,
          'action' => 3,
          'calInvite' => false,
          'uid' => intval($row['user_id']),
        ];
      }
      $triggerResult = $this->phpbb_dispatcher->trigger_event('msta.emailontopic.receivers_prepare_list', array(
        'list' => $mailtolist,
        'postId' => $event['post_id'],
      ));
      if ($triggerResult && gettype($triggerResult['list']) == 'array') $mailtolist = $triggerResult['list'];
      foreach ($mailtolist as $val) {
        if ($list != '') $list .= '|';
        $list .= urlencode($val['name']).';'.(($val['isGroup'])?'g':'u').';'.$val['action'].';'.(($val['isNew'])?'n':'d').';'.(($val['calInvite'])?'i':'n');
      }
    }
    // assign vars
    /*
     $list format: multiple elements separeated by |, each element consists of
     user/groupname
     u/g: for user/group
     mode: 0=delete  1=normal email  2=normal subscription  3=enduring subscription
     n/d/D: for new element / database enduring subscription with (d) or without (D) permission to be deleted
    */
    $this->template->assign_vars([
      'EMAIL_ENABLED' => $mailEnabled,
      'EMAIL_LIST' => $list,
      'CHOSEN_TELEGRAM_CHANNELS' => $telegramChannels,
      'CAN_CONFIG_TELEGRAM' => (($this->canConfigTelegram())?1:0),
    ]);
  }

  private function canConfigTelegram () {
    $postId = intval($this->request->variable('p', '', true));
    $topicId = $this->getTopicId();
    $forumId = intval($this->request->variable('f', '', true));
    // ensure, user has the permission to configure telegram links
    if ($forumId == 0) return false;
    if (!$this->auth->acl_get('f_configtelegram', $forumId)) return false;
    // ensure, this is the "first/main post" in the topic
    if ($topicId == 0) return true;
    $firstPostRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id FROM '.$this->table_prefix.'posts WHERE topic_id = '.$topicId.' ORDER BY post_time LIMIT 1'));
    if ($firstPostRow && $firstPostRow['post_id'] != $postId) return false;
    return true;
  }

  private function setTelegramLinks ($event) {
    $this->request->enable_super_globals();
    $topicId = $this->getTopicId();
    $postId = intval($this->request->variable('p', '', true));
    $forumId = intval($this->request->variable('f', '', true));
    if (!$this->canConfigTelegram()) return;
    // unlink all existing channels from this topic
    $this->db->sql_query('UPDATE '.$this->table_prefix.'telegram_links SET forum_link_id = 0, link_type = 0 WHERE forum_link_id = '.$topicId);
    // link all channels from request with this topic
    $postTelegramCh = $_POST['telegramChannels'];
    if ($postTelegramCh) foreach ($postTelegramCh as $val) {
      $spl = explode('|', $val);
      if (count($spl) < 2) $spl[1] = 3;
      $spl[0] = intval($spl[0]);
      $spl[1] = intval($spl[1]);
      if ($spl[1] > 3 || $spl[1] <= 0) $spl[1] = 3;
      $this->db->sql_query('UPDATE '.$this->table_prefix.'telegram_links SET forum_link_id = '.$topicId.', link_type = '.$spl[1].' WHERE telegram_chat_id = '.$spl[0]);
    }
  }

  private function processUGList ($event) {
    $mailtolist = array();
    $subscriptionlist = array();
    $spl = explode('|', $this->request->variable('mailSubscriptions', '', true));
    for ($i = 0; $i < count($spl); $i++) {
      $spl[$i] = explode(';', $spl[$i]);
      /*
       spl[i][0] = name of user or group (uri-escaped)
       spl[i][1] = u or g for user or group
       spl[i][2] = 0:remove subscription  1:one time message  2:message+set std subscription  3:set enduring subscription
       spl[i][3] = d or n for database existing entry or new entry
       spl[i][4] = i or n or '' if the user or group should be invited to the post's event (calendar-extension)
      */
      if (count($spl[$i]) < 5) continue;
      $mailtolist[] = [
        'name' => urldecode($spl[$i][0]),
        'isNew' => $spl[$i][3] == 'n',
        'isGroup' => $spl[$i][1] == 'g',
        'action' => intval($spl[$i][2]),
        'calInvite' => $spl[$i][4] == 'i',
      ];
    }
    return $mailtolist;
  }

  public function input_bridge_post_end ($event) {
    if (!$this->bridgeManager)
      $this->bridgeManager = new \msta\emailontopic\includes\BridgeManager($this->phpbb_container, $this->db, $this->config, $this->user, $this->phpbb_dispatcher, $this->language, $this->auth, $this->phpbb_root_path);
    $err = $this->bridgeManager->post_bridging_trigger($event);
    $event['err'] = $err;
  }

  public function submit_post_end ($event) {
    // store bridge settings
    $this->setTelegramLinks($event);
    // trigger bridges
    $bbcode_options = (($event['data']['enable_bbcode']) ? OPTION_FLAG_BBCODE : 0) +
                      (($event['data']['enable_smilies']) ? OPTION_FLAG_SMILIES : 0) + 
                      (($event['data']['enable_urls']) ? OPTION_FLAG_LINKS : 0);
    $bridgeEvent = array(
      'input' => 'forum',
      'mode' => $event['mode'],
      'forumId' => intval($event['data']['forum_id']),
      'topicId' => intval($event['data']['topic_id']),
      'postId' => intval($event['data']['post_id']),
      'message' => generate_text_for_display($event['data']['message'], $event['data']['bbcode_uid'], $event['data']['bbcode_bitfield'], $bbcode_options),
      'topicTitle' => $event['data']['topic_title'],
      'postTitle' => $this->request->variable('subject', $event['data']['topic_title'], true),
      'mailToList' => $this->processUGList($event),
    );
    if (!$this->bridgeManager)
      $this->bridgeManager = new \msta\emailontopic\includes\BridgeManager($this->phpbb_container, $this->db, $this->config, $this->user, $this->phpbb_dispatcher, $this->language, $this->auth, $this->phpbb_root_path);
    $err = $this->bridgeManager->post_bridging_trigger($bridgeEvent);
    // write conclusion to cookie
    $cookie = '';
    foreach ($err as $k => $v) {
      if (count($v) > 0) {
        if ($cookie != '') $cookie .= '|';
        $cookie .= $k;
      }
      for ($i = 0; $i < count($v); $i++)
        $cookie .= '#'.rawurlencode($v[$i]);
    }
    setrawcookie('EMAILONTOPIC', $cookie);
  }

  public function delete_posts ($event) {
    if (!$this->bridgeManager)
      $this->bridgeManager = new \msta\emailontopic\includes\BridgeManager($this->phpbb_container, $this->db, $this->config, $this->user, $this->phpbb_dispatcher, $this->language, $this->auth, $this->phpbb_root_path);
    if ($this->config['telegram_webhook_pass'])
      for ($i = 0; $i < count($event['post_ids']); $i++) {
        $this->bridgeManager->delete_bridging_trigger(array(
          'input' => 'forum',
          'forumId' => intval($event['forum_ids'][$i]),
          'topicId' => intval($event['topic_ids'][$i]),
          'postId' => intval($event['post_ids'][$i]),
        ));
      }
  }
}



