<?php
/**
*
* @package msta EmailOnTopic Extension
* This class manages the creation of alle requires bridge interfaces and searches and composes all users to be notified.
*
*/

namespace msta\emailontopic\includes;

namespace msta\emailontopic\includes;

class BridgeManager {
  protected $phpbb_container;
  protected $db;
  protected $config;
  protected $user;
  protected $phpbb_dispatcher;
  protected $language;
  protected $auth;
  protected $phpbb_root_path;
  private $table_prefix;
  private $userQuery;

  private $template = null;
  private $permissionCheck = null;

  private $availlableBridges = array( // ToDo: Create and cache this array automatically
    'telegram' => array(
      'class' => '\msta\emailontopic\includes\bridges\TelegramBridge',
      'sendSubscriptionDefault' => false,
    ),
    'email' => array(
      'class' => '\msta\emailontopic\includes\bridges\MailMessenger',
      'sendSubscriptionDefault' => true,
    ),
  );

  public function __construct($phpbb_container, \phpbb\db\driver\driver_interface $db, \phpbb\config\config $config, \phpbb\user $user, \phpbb\event\dispatcher_interface $phpbb_dispatcher, \phpbb\language\language $language, \phpbb\auth\auth $auth, $phpbb_root_path) {
    $this->phpbb_container = $phpbb_container;
    $this->db = $db;
    $this->config = $config;
    $this->user = $user;
    $this->phpbb_dispatcher = $phpbb_dispatcher;
    $this->language = $language;
    $this->auth = $auth;
    $this->phpbb_root_path = $phpbb_root_path;
    $this->table_prefix = substr(USERS_TABLE, 0, -5);
    $this->userQuery = 'SELECT u.user_id, u.username, u.user_email, u.user_jabber, u.user_lang, u.mail_access_password, pf.pf_pgp_fingerprint FROM '.$this->table_prefix.'users AS u LEFT JOIN '.$this->table_prefix.'profile_fields_data AS pf ON u.user_id = pf.user_id WHERE (u.user_type = '.USER_NORMAL.' OR u.user_type = '.USER_FOUNDER.')';
  }

  private function setupTemplate () {
    $template_environment = new \phpbb\template\twig\environment(
      $this->phpbb_container->get('config'),
      $this->phpbb_container->get('filesystem'),
      $this->phpbb_container->get('path_helper'),
      $this->phpbb_container->getParameter('core.template.cache_path'),
      $this->phpbb_container->get('ext.manager'),
      new \phpbb\template\twig\loader($this->phpbb_container->get('filesystem')),
      $this->phpbb_dispatcher,
      array()
    );
    $template_environment->setLexer($this->phpbb_container->get('template.twig.lexer'));
    $this->template = new \phpbb\template\twig\twig(
      $this->phpbb_container->get('path_helper'),
      $this->phpbb_container->get('config'),
      new \phpbb\template\context(),
      $template_environment,
      $this->phpbb_container->getParameter('core.template.cache_path'),
      null,
      $this->phpbb_container->get('template.twig.extensions.collection'),
      $this->phpbb_container->get('ext.manager')
    );
    $this->template->set_custom_style(
      array(array(
        'name' 		=> 'de' . '_email',
        'ext_path' 	=> 'language/' . 'de' . '/email',
      )),
      array(
        $this->phpbb_root_path.'ext/msta/emailontopic/styles/messenger/',
      )
    );
    $this->template->set_filenames(array(
      'html_chat_forwarding' => 'html_chat_forwarding.html',
      'html_post_notification' => 'html_post_notification.html',
      'html_edit_post_info' => 'html_edit_post_info.html',
    ));
  }



  /*
   getAndMarkSubscribers
   returns an associative array (U+id as key) of all users, who should be notified due to subscriptions. Sets the notify_status to 1 for topic subscriptions.
  */
  private function getAndMarkSubscribers ($postMode, $forumId, $topicId) {
    $users = array();
    $subscribedusers = null;
    if ($postMode == 'post')
      $subscribedusers = $this->db->sql_query('SELECT user_id FROM '.$this->table_prefix.'forums_watch'.' WHERE (forum_id = '.intval($forumId).' AND notify_status = 0)');
    else if ($postMode != 'edit')
      $subscribedusers = $this->db->sql_query('SELECT user_id, enduring FROM '.$this->table_prefix.'topics_watch'.' WHERE (topic_id = '.intval($topicId).' AND (notify_status = 0 OR enduring = 1))');
    $subscribedIds = array();
    if ($subscribedusers)
      foreach ($subscribedusers as $subscriber) {
        $subscribedIds[] = intval($subscriber['user_id']);
        if ($this->user->data['user_id'] == $subscriber['user_id']) continue; // don't notify posting user
        $userRow = $this->db->sql_fetchrow($this->db->sql_query($this->userQuery.' AND u.user_id = '.$subscriber['user_id']));
        if ($userRow)
          $users['U'.intval($subscriber['user_id'])] = array(
            'name'		=> $userRow['username'],
            'email'		=> $userRow['user_email'],
            'jabber'	=> $userRow['user_jabber'],
            'lang'		=> $userRow['user_lang'],
            'reason'	=> ($postMode == 'post')?'forum_subscribe':'topic_subscribe'.((intval($subscriber['enduring']) == 1)?'_enduring':''),
            'notiType'	=> ($postMode == 'post')?'topic':'post',
            'noti'		=> null, // initialized later!
            'uid'		=> intval($subscriber['user_id']),
            'pwd'		=> $userRow['mail_access_password'],
            'pgp'		=> $userRow['pf_pgp_fingerprint'],
          );
      }
    if ($postMode != 'post' && count($subscribedIds) > 0)
      $this->db->sql_query('UPDATE '.$this->table_prefix.'topics_watch'.' SET notify_status = 1 WHERE (user_id IN ('.join(', ', $subscribedIds).') AND topic_id = '.intval($topicId).')');
    return $users;
  }

  private function getDirectAddressedUsersAndGroups (&$users, &$groups, &$doSubscriptionManagement, $mailToList) {
    if (!$mailToList) return array();
    // prepare sql queries
    $sqlUsers = array();
    $sqlGroups = array();
    for ($i = 0; $i < count($mailToList); $i++)
      if ($mailToList[$i]['isGroup']) $sqlGroups[] = "'".$this->db->sql_escape($mailToList[$i]['name'])."'";
      else $sqlUsers[] = "'".$this->db->sql_escape($mailToList[$i]['name'])."'";
    // fetch users from db and add existing ones to given user list
    if (count($sqlUsers) > 0) {
      $sqlRes = $this->db->sql_query($this->userQuery.' AND u.username IN ('.join(', ', $sqlUsers).')');
      while ($row = $this->db->sql_fetchrow($sqlRes)) {
        $i = 0;
        while ($i < count($mailToList))
          if ($mailToList[$i]['name'] == $row['username']) break;
          else $i++;
        if ($i >= count($mailToList)) continue;
        if ($mailToList[$i]['action'] != 1) $doSubscriptionManagement = true;
        $users['U'.intval($row['user_id'])] = array_merge($mailToList[$i], array(
          'name'		=> $row['username'],
          'email'		=> $row['user_email'],
          'jabber'		=> $row['user_jabber'],
          'lang'		=> $row['user_lang'],
          'reason'		=> 'user_explicit_notify',
          'notiType'	=> 'explicit',
          'noti'		=> null,
          'uid'			=> intval($row['user_id']),
          'pwd'			=> $row['mail_access_password'],
          'pgp'			=> $row['pf_pgp_fingerprint'],
        ));
      }
    }
    // fetch groups from db and add existing ones to given user list
    if (count($sqlGroups) > 0) {
      $sqlRes = $this->db->sql_query('SELECT group_id, group_name FROM '.$this->table_prefix.'groups'.' WHERE group_name IN ('.join(', ', $sqlGroups).')');
      while ($row = $this->db->sql_fetchrow($sqlRes)) {
        $i = 0;
        while ($i < count($mailToList))
          if ($mailToList[$i]['name'] == $row['group_name']) break;
          else $i++;
        if ($i >= count($mailToList)) continue;
        $groups[] = array_merge($mailToList[$i], array(
          'name'	=> $row['group_name'],
          'gid'	=> intval($row['group_id']),
        ));
      }
    }
    // subscription management
    $notExisting = array();
    for ($i = 0; $i < count($mailToList); $i++)
      if ($mailToList[$i]['uid'] || $mailToList[$i]['gid']) $notExisting[] = (($mailToList[$i]['isGroup'])?'*':'').$mailToList[$i]['name'];
    return $notExisting;
  }

  private function manageSubscriptions (&$directUsers, $tid) {
    foreach ($directUsers as $key => $value)
      if ($value['action'] != 1 && $this->permissionCheck->canManageSubscriptions($value['uid'])) {
        if ($value['action'] == 2 || $value['action'] == 3 && $value['isNew'])
          $this->permissionCheck->setSubscription(intval($value['uid']), $tid, intval($value['action'])-1);
        else if (intval($spl[$i][2]) == 0)
          $this->permissionCheck->setSubscription(intval($value['uid']), $tid, -1);
      }
  }

  private function addGroupUsers ($directGroups, &$users) {
    $missingPermissions = array();
    $adressableGroups = array();
    $groupMap = array();
    for ($i = 0; $i < count($directGroups); $i++) {
      if ($directGroups[$i]['action'] == 0) continue;
      if ($this->permissionCheck->canAddressGroup($directGroups[$i]['gid'])) $adressableGroups[] = intval($directGroups[$i]['gid']);
      else $missingPermissions[] = '*'.$directGroups[$i]['name'];
    }
    $groupUsers = array();
    if (count($adressableGroups) > 0) {
      $sqlRes = $this->db->sql_query('SELECT user_id, group_id FROM '.$this->table_prefix.'user_group WHERE group_id IN ('.join(', ', $adressableGroups).') AND user_pending = 0');
      while ($row = $this->db->sql_fetchrow($sqlRes)) {
        $groupUsers[] = intval($row['user_id']);
        for ($i = 0; $i < count($directGroups); $i++)
          if ($directGroups[$i]['gid'] == $row['group_id'])
            $groupMap['U'.$row['user_id']] = $directGroups[$i]['name'];
      }
    }
    if (count($groupUsers) > 0) {
      $sqlRes = $this->db->sql_query($this->userQuery.' AND u.user_id IN ('.join(', ', $groupUsers).')');
      while ($row = $this->db->sql_fetchrow($sqlRes)) {
        $users['U'.intval($row['user_id'])] = array(
          'name'		=> $row['username'],
          'email'		=> $row['user_email'],
          'jabber'		=> $row['user_jabber'],
          'lang'		=> $row['user_lang'],
          'reason'		=> 'group_notify',
          'notiType'	=> 'group',
          'noti'		=> null,
          'uid'			=> intval($row['user_id']),
          'pwd'			=> $row['mail_access_password'],
          'pgp'			=> $row['pf_pgp_fingerprint'],
          'gname'		=> $groupMap['U'.$row['user_id']],
        );
      }
    }
    return $missingPermissions;
  }

  private function addAddressableUsers($directUsers, &$users) {
    $mailOverLimit = array();
    $nrDirectAddressed = 0;
    foreach ($directUsers as $k => $v) {
      if ($v['action'] == 0) continue;
      $limit = $this->permissionCheck->addressingLimitForUser($v['uid']);
      if ($limit > $nrDirectAddressed) {
        $users[$k] = $v;
        $nrDirectAddressed++;
      } else if (!array_key_exists($k, $users)) $mailOverLimit[] = $v['name'].' (max. '.$limit.')';
    }
    return $mailOverLimit;
  }

  /*
   getNotificationSettings
   sets for a given array $users the noti-Array, holding the information, to which bridges the notification sould be applied.
   Returns an array with information, which bridges are required in general.
  */
  private function getNotificationSettings (&$users, &$requiredBridges) {
    $deactivateCnt = array();
    foreach ($requiredBridges as $k => $v)
      $deactivateCnt[$k] = 0;
    $uids = array();
    foreach ($users as $k => $val) {
      $uids[] = $val['uid'];
      $users[$k]['noti'] = $requiredBridges; // clones the array!
    }
    if (count($uids) > 0) {
      $notiSql = $this->db->sql_query('SELECT item_type, user_id, method, notify FROM '.$this->table_prefix.'user_notifications WHERE user_id IN ('.join(', ', $uids).')');
      while ($row = $this->db->sql_fetchrow($notiSql)) {
        $bridge = array_pop(explode('.', $row['method']));
        $notiType = array_pop(explode('.', $row['item_type']));
        if (array_key_exists($bridge, $requiredBridges) && $users['U'.intval($row['user_id'])]['notiType'] == $notiType)
          if ($row['notify'] == 1) {
            $users['U'.intval($row['user_id'])]['noti'][$bridge] = true;
            $requiredBridges[$bridge] = true;
          } else
            $deactivateCnt[$bridge]++;
      }
    }
    foreach ($deactivateCnt as $k => $v)
      if ($v == count($users))
        $requiredBridges[$k] = false;
  }

  /*
   getForwardingBridges
   sets the entry in the given requiredBridges-array to true, if a topic-forwarding to this bridge is active for the given topicId.
   Returns true, if at least one bridge has an active topic-forwarding.
  */
  private function getForwardingBridges ($topicId, &$requiredBridges) {
    $forwardingBridges = false;
    foreach ($this->availlableBridges as $key => $bridgeInfo)
      if ($bridgeInfo['class']::linkedChannelExists($topicId, $this->db, $this->table_prefix)) {
        $requiredBridges[$key] = true;
        $forwardingBridges = true;
      }
    return $forwardingBridges;
  }

  /*
   generateMailAccessPWD
   Checks, if a given password is empty or too short. If this is the case, this function generates a random password with 12 chars and stores it in the database. The password is returned as a var parameter.
  */
  private function generateMailAccessPWD (&$pwd, $userId) {
    if (!$pwd || strlen($pwd) < 12) {
      $pwd = '';
      for ($i = 0; $i < 12; $i++) {
        $c = mt_rand(0, 61);
        if ($c < 10) $pwd .= chr($c+48);
        else if ($c < 36) $pwd .= chr($c-10+65);
        else $pwd .= chr($c-36+97);
      }
      $this->db->sql_query('UPDATE '.$this->table_prefix."users SET mail_access_password = '".$pwd."' WHERE user_id = ".intval($userId));
    }
  }

  /*
   generateTopicPathAndSecurity
   Iterates upwards through the forum tree begining with the given forumID and generates a path of forums as a string with each forum seperated with a /.
   Result: The return value is a variable, holding the cumulated security flags of the forums. (Rule: A forum security restriction holds for all sub-forums and cannot be weakened by sub-forum settings)
  */
  private function generateTopicPathAndSecurity (&$topicPath, $forumId) {
    $security = 0;
    $nextFid = intval($forumId);
    do {
      $forumRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT forum_id, parent_id, forum_name, bridging_security FROM '.$this->table_prefix.'forums WHERE forum_id = '.$nextFid));
      $nextFid = 0;
      if ($forumRow) {
        $topicPath = $forumRow['forum_name'].'/'.$topicPath;
        $nextFid = $forumRow['parent_id'];
        if ((intval($forumRow['bridging_security']) & 96) > ($security & 96))
          $security = ($security & ~96) | (intval($forumRow['bridging_security']) & 96);
        $security = $security | (intval($forumRow['bridging_security']) & 31);
      }
    } while ($nextFid != 0);
    return $security;
  }

  /*
   completeEventInfo
   The given $event (array or event-object) will be completed with missing info from the db.
   If $content is true, message content, title etc. is completed, if false, only the forum- and topic-IDs.
  */
  private function completeEventInfo (&$event, $content) {
    if (!$event['forumId'] || !$event['topicId'] || $content && (gettype($event['message']) != 'string' || gettype($event['topicTitle']) != 'string' || gettype($event['postTitle']) != 'string')) {
      $infoRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT p.topic_id, p.forum_id, p.post_subject, p.post_text, t.topic_title FROM '.$this->table_prefix.'posts AS p INNER JOIN '.$this->table_prefix.'topics AS t ON p.topic_id = t.topic_id WHERE p.post_id = '.intval($event['postId'])));
      if (!$infoRow) return false;
      $event['forumId'] = intval($infoRow['forum_id']);
      $event['topicId'] = intval($infoRow['topic_id']);
      if ($content) {
        $event['message'] = $infoRow['post_text'];
        $event['topicTitle'] = $infoRow['topic_title'];
        $event['postTitle'] = $infoRow['post_subject'];
      }
    }
    if (gettype($event['inputInfo']) == 'integer') $event['inputInfo'] = strval($event['inputInfo']);
    if (gettype($event['inputInfo']) != 'string') $event['inputInfo'] = '';
    return true;
  }


  /*
   post_bridging_trigger
   This function spreads new posts to the different bridges. Called (directly or via event input_bridge_post_end), when a new post is generated, after is is stored in the forum's database.
   $event: an array or \phpbb\event\data containing
    - input: The name of the bridge, that created the post ('forum' if directly written in phpBB)
    - mode: the mode of the post (i.e. 'post', 'edit', 'reply')
    - postId: the id of the new post
    optional:
    - inputInfo: String with additional information generated by the input bridge, for example to identify the exact input channel id
    - forumId: the id of the forum, in which the post was created -> if not set, fetched from database
    - topicId: the id of the topic, in which the post was created -> if not set, fetched from database
    - message: the HTML-formatted message, ready to be sent to the database (ensure before, that no forbidden entities are included) -> if not set, fetched from database
    - topicTitle: the title of the topic -> if not set, fetched from database
    - postTitle: the title of the post -> if not set, fetched from database
    - author: the author of the post -> if not set, current user taken
    - mailtolist: when input is the forum, the list of target users and groups to be notified explicitly
   result: array of errors (each error is a pair of array key and value; value is again an array of detailed information). Empty if no error occured.
  */
  public function post_bridging_trigger ($event) {
    // load missing values from database
    if (!$this->completeEventInfo($event, true))
      return array('GENERAL_ERROR' => array('Incomplete event info!'));
    if (!array_key_exists('mailtolist', $event))
      $event['mailtolist'] = null;
    // init permission check
    $this->permissionCheck = \msta\emailontopic\includes\PermissionCheck::getInstance($this->db, $this->auth);
    // get users and groups to notify due to subscriptions or explicit notifications
    $err = array(
      'SUBS_MISSING_PERMISSIONS' => array(),
      'DISABLED_BRIDGES' => array(),
    );
    $users = $this->getAndMarkSubscribers($event['mode'], $event['forumId'], $event['topicId']);
    $directUsers = array();
    $directGroups = array();
    $doSubscriptionManagement = false;
    $err['SUBS_NON_EXISTING'] = $this->getDirectAddressedUsersAndGroups($directUsers, $directGroups, $doSubscriptionManagement, $event['mailToList']);
    $this->permissionCheck->initAddressingPermissions($this->user->data['user_id'], $directUsers, $directGroups, $doSubscriptionManagement);
    $this->manageSubscriptions($directUsers, intval($event['topicId']));
    $this->phpbb_dispatcher->trigger_event('msta.emailontopic.receivers_receive_list', array(
      'users' => $directUsers,
      'groups' => $directGroups,
      'postId' => $event['postId'],
    ));
    if ($triggerResult && gettype($triggerResult['message']) == 'string') $message = $triggerResult['message'];
    $err['SUBS_GROUP_MISSING_PERMISSIONS'] = $this->addGroupUsers($directGroups, $users);
    $err['SUBS_OVER_LIMIT'] = $this->addAddressableUsers($directUsers, $users);
    $directUsers = null;
    $directGroups = null;
    $this->permissionCheck->clearAddressingPermissions();
    // compose information in required bridges
    $requiredBridges = array();
    foreach ($this->availlableBridges as $key => $bridgeInfo)
      $requiredBridges[$key] = $bridgeInfo['sendSubscriptionDefault'];
    $this->getNotificationSettings($users, $requiredBridges);
    $forwardingBridges = $this->getForwardingBridges($event['topicId'], $requiredBridges);
    // create all used bridges
    $bridges = array();
    foreach ($this->availlableBridges as $key => $bridgeInfo)
      if ($requiredBridges[$key]) {
        $bridges[$key] = $bridgeInfo['class']::getInstance($this->phpbb_container, $this->db, $this->config, $this->phpbb_dispatcher, $this->phpbb_root_path);
        if (!$bridges[$key])
          $err['DISABLED_BRIDGES'][] = $key;
      } else $bridges[$key] = null;
    // trigger send mail event
    $message = $event['message'];
    $triggerResult = $this->phpbb_dispatcher->trigger_event('msta.emailontopic.send_bridges', array(
      'forumId' => $event['forumId'],
      'topicId' => $event['topicId'],
      'postId'  => $event['postId'],
      'message' => $message,
      'mail_access_triggers' => array(),
      'vars' => array('templates_before_message' => array()),
    ));
    if ($triggerResult && gettype($triggerResult['message']) == 'string') $message = $triggerResult['message'];
    // set fix template vars
    $this->setupTemplate();
    $this->user->add_lang_ext('msta/emailontopic', 'messenger');
    $topicPath = $event['topicTitle'];
    $security = $this->generateTopicPathAndSecurity($topicPath, $event['forumId']);
    $this->template->assign_vars(array_merge($triggerResult['vars'], array(
      'mode'			=> $event['mode'],
      'author'			=> (array_key_exists('author', $event))?$event['author']:$this->user->data['username'],
      'posturl'			=> generate_board_url().'/viewtopic.php?f='.$event['forumId'].'&t='.$event['topicId'].'&p='.$event['postId'].'#p'.$event['postId'],
      'authorurl'		=> ($event['input'] == 'forum')?generate_board_url().'/memberlist.php?mode=viewprofile&u='.intval($this->user->data['user_id']):'',
      'ucpSettings'		=> generate_board_url().'/ucp.php?i=ucp_prefs&mode=personal',
      'topic'			=> $topicPath,
      'boardSig'		=> $this->config['board_email_sig'],
      'noContent'		=> (($security & 96) == 96)?1:0,
      'allowClearMsg'	=> ((($security & 96) <= 32))?1:0,
      'isMessangerTpl'	=> 1,
    )));
    // send forwardings
    if ($forwardingBridges) {
      // set lang to basename($config['default_lang'])
      foreach ($bridges as $key => $bridge)
        if ($bridge) {
          $lastErr = $bridge->sendChannels(
            $event['input'],
            $event['inputInfo'],
            $event['mode'],
            $this->template,
            $message,
            $event['forumId'],
            $event['topicId'],
            $event['postId']
          );
          if ($lastErr && count($lastErr) > 0)
            if (!array_key_exists('CHANNEL_POST_ERRORS', $err)) $err['CHANNEL_POST_ERRORS'] = $lastErr;
            else $err['CHANNEL_POST_ERRORS'] = array_merge($err['CHANNEL_POST_ERRORS'], $lastErr);
        }
    }
    // send subscriptions
    foreach ($users as $k => $val)
      if ($this->permissionCheck->userCanReadForum($val['uid'], $event['forumId'])) {
        $this->generateMailAccessPWD($val['pwd'], $val['uid']);
        $this->language->set_user_language($val['lang'], true);
        $triggerBegin = generate_board_url().'/app.php/emailontopic/getinfo?p='.$event['postId'].'&u='.$val['uid'].'&pwd='.rawurlencode($val['pwd']);
        $triggers = array();
        foreach ($triggerResult['mail_access_triggers'] as $triggerKey => $triggerVal)
          $triggers[$triggerKey] = $triggerBegin.'&action=trigger&key='.$triggerKey.(($triggerVal == '')?'':'&val='.rawurlencode($triggerVal));
        $this->template->assign_vars(array(
          'reason'			=> $val['reason'],
          'allowClearMsg'	=> (((($security & 96) == 0) || ((($security & 96) == 32) && ($val['reason'] == 'group_notify' || $val['reason'] == 'user_explicit_notify')))?1:0),
          'group'			=> $val['gname'],
          'postinfo'		=> $triggerBegin.'&action=mailnewerpostinfo&f='.$event['forumId'].'&t='.$event['topicId'],
          'unsubscribeurl'	=> $triggerBegin.'&action=unsubscribe&t='.$event['topicId'],
          'triggers'		=> $triggers,
        ));
        foreach ($val['noti'] as $bridge => $enabled)
          if ($enabled) {
            if (!array_key_exists($bridge, $bridges) || !$bridges[$bridge]) continue;
            $success = $bridges[$bridge]->sendSubscription(
              $event['input'],
              $event['inputInfo'],
              $val,
              $this->template,
              $message,
              (($security & 1) == 1 || ($security & 96) == 96)?$this->language->lang('EMAILONTOPIC_SUBJECT'):$event['postTitle'],
              (($security & 96) == 96)?$this->language->lang('EMAILONTOPIC_SUBJECT'):$event['postTitle'],
              $val['reason'],
              ($event['input'] == 'forum')?$this->user:null
            );
            if ($success) {
              if (!array_key_exists('SUBS_SEND_SUCCESS', $err)) $err['SUBS_SEND_SUCCESS'] = array();
              $err['SUBS_SEND_SUCCESS'][] = $val['name'];
            } else {
              if (!array_key_exists('SUBS_SEND_FAIL', $err)) $err['SUBS_SEND_FAIL'] = array();
              $err['SUBS_SEND_FAIL'][] = $val['name'];
            }
          }
      } else $err['SUBS_MISSING_PERMISSIONS'][] = $val['name'];
    foreach ($bridges as $b)
      if ($b)
        $b->finalSteps();
    return $err;
  }

  /*
   delete_bridging_trigger
   This function spreads the deletion of a post to all bridges. Should be called, before post is removed from forum's database!
   $event: an array or \phpbb\event\data containing
    - input: The name of the bridge, that caused the deletion
    - postId: the id of the post to be deleted
    optional:
    - forumId: the id of the forum
    - topicId: the id of the topic
    - reason: the reason, why the post was deleted or moved
   result: array of errors (each error is a pair of array key and value; value is again an array of detailed information). Empty if no error occured.
  */
  public function delete_bridging_trigger ($event) {
    if (!$this->completeEventInfo($event, false)) // TODO: For unknown reason, completeEventInfo fails here. Just ignore this for the moment and ensure to use full $event info
      return array('GENERAL_ERROR' => array('Incomplete event info!'));
    $err = array(
      'DISABLED_BRIDGES' => array(),
      'DELETE_SUCCESS' => array(),
      'DELETE_FAIL' => array(),
    );
    // compose information in required bridges
    $requiredBridges = array();
    foreach ($this->availlableBridges as $key => $bridgeInfo)
      $requiredBridges[$key] = $bridgeInfo['sendSubscriptionDefault'];
    if (!$this->getForwardingBridges($event['topicId'], $requiredBridges))
      return; // no bridge has a forwarding, that supports deletion of posts
    // create bridges and call delete trigger
    foreach ($this->availlableBridges as $key => $bridgeInfo)
      if ($requiredBridges[$key]) {
        $bridge = $bridgeInfo['class']::getInstance($this->phpbb_container, $this->db, $this->config, $this->phpbb_dispatcher, $this->phpbb_root_path);
        if ($bridge) {
          $currentErr = $bridge->deletePost(
            $event['input'],
            $event['inputInfo'],
            $event['forumId'],
            $event['topicId'],
            $event['postId'],
            (array_key_exists('reason', $event))?strval($event['reason']):''
          );
          if (is_array($currentErr) && count($currentErr) > 0)
            $err['DELETE_FAIL'] = array_merge($err['DELETE_FAIL'], $currentErr);
          else
            $err['DELETE_SUCCESS'][] = $key;
        } else $err['DISABLED_BRIDGES'][] = $key;
      }
    return $err;
  }

}
