<?php
/**
*
* @package msta EmailOnTopic Extension
*
*/

namespace msta\emailontopic\includes;

class PermissionCheck {
  protected $db;
  protected $auth;
  private $groups = array(); // the groups of each user (cache)
  private $table_prefix;
  private $user_group = null;
  private $permissions = null;
  private $allowCompleteGroup = 255;
  private static $instance = null;

  private function __construct () { }
  public function getInstance (\phpbb\db\driver\driver_interface $db, \phpbb\auth\auth $auth) {
    if (self::$instance == null) {
      self::$instance = new PermissionCheck();
      self::$instance->db = $db;
      self::$instance->auth = $auth;
      self::$instance->table_prefix = substr(USERS_TABLE, 0, -5);
    }
    return self::$instance;
  }
  public function getExistingInstance () {
    return self::$instance;
  }



  public function initAddressingPermissions ($uid, $users, $groups, $doSubscriptionManagement = false) {
    if (count($groups) == 0 && !$doSubscriptionManagement) { // the fast way: test default permission of registered users first and use only them if sufficient
      $permissionRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT user_recipient_count FROM '.$this->table_prefix.'group_mail_privileges WHERE from_group_id = 2 AND to_group_id = 2'));
      if ($permissionRow && intval($permissionRow['user_recipient_count']) >= count($users)) {
        $this->user_group = null;
        $this->permissions = array(array(
          'from_group_id'			=> 2,
          'to_group_id'				=> 2,
          'user_recipient_count'	=> $permissionRow['user_recipient_count'],
          'subscription_management' => 0,
        ));
        return;
      }
    }
    // get list of all groups of all involved users
    $involvedUserIDs = array(intval($uid));
    foreach ($users as $k => $v) $involvedUserIDs[] = $v['uid'];
    $this->user_group = $this->db->sql_query('SELECT group_id, user_id, group_leader FROM '.$this->table_prefix.'user_group WHERE user_id IN ('.join(', ', $involvedUserIDs).') AND user_pending = 0');
    // get list of all permissions of all involved sender -> recipient group relations
    $senderGroups = array();
    $recipientGroups = array();
    foreach ($groups as $group)
      $recipientGroups[] = intval($group['gid']);
    foreach ($this->user_group as $k => $row) {
      if ($row['user_id'] == $uid)
        $senderGroups[] = intval($row['group_id']);
      $recipientGroups[] = intval($row['group_id']);
    }
    // load permissions and transform from a sql response to a true array
    $this->permissions = array();
    if (count($senderGroups) > 0 && count($recipientGroups) > 0) {
      $this->permissions_response = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'group_mail_privileges WHERE (from_group_id IN ('.join(', ', $senderGroups).')) AND (to_group_id IN('.join(', ', $recipientGroups).'))');
      while ($row = $this->db->sql_fetchrow($this->permissions_response)) {
        $newV = array();
        foreach ($row as $k => $v) $newV[$k] = $v;
        $this->permissions[] = $newV;
      }
    }
    // overwrite user-permissions with leader-permissions, if $uid is group-leader
    foreach ($this->user_group as $k => $row)
      if ($row['user_id'] == $uid && intval($row['group_leader']) == 1)
        for ($i = 0; $i < count($this->permissions); $i++)
          if ($this->permissions[$i]['from_group_id'] == $row['group_id']) {
            if ($this->permissions[$i]['user_recipient_count'] < $this->permissions[$i]['leader_recipient_count'])
              $this->permissions[$i]['user_recipient_count'] = $this->permissions[$i]['leader_recipient_count'];
            if ($this->permissions[$i]['subscription_management'] == 1)
              $this->permissions[$i]['subscription_management'] = 2;
          }
  }

  /*
   canManageSubscriptions
   Returns true, if the user given to getAddressingPermissions is allowed to manage subscriptions of $uid
  */
  public function canManageSubscriptions ($uid) {
    if ($this->user_group === null) throw 'Not initialized for management!';
    if ($this->permissions === null) throw 'Adressing permissions not initialized!';
    foreach ($this->user_group as $vUG)
      if ($vUG['user_id'] == $uid)
        foreach ($this->permissions as $vP)
          if ($vP['to_group_id'] == $vUG['group_id'] && $vP['subscription_management'] == 2) // this woks, because user permissions are overwritten with leader permissions
            return true;
    return false;
  }

  /*
   addressingLimitForUser
   Returns the maximum number of recipients, that are allowed for the constellation of groups in which the $uid is member
  */
  public function addressingLimitForUser ($uid) {
    if ($this->permissions === null) throw 'Adressing permissions not initialized!';
    if ($this->user_group === null) return intval($this->permissions[0]['user_recipient_count']); // this is the case, if the fast approach is used (see beginning of initAddressingPermissions)
    $maxAddressable = 0;
    foreach ($this->user_group as $vUG)
      if ($vUG['user_id'] == $uid)
        foreach ($this->permissions as $vP)
          if ($vP['to_group_id'] == $vUG['group_id'] && intval($vP['user_recipient_count']) > $maxAddressable)
            $maxAddressable = intval($vP['user_recipient_count']);
    return $maxAddressable;
  }

  public function canAddressGroup ($gid) {
    if ($this->permissions === null) throw 'Adressing permissions not initialized!';
    foreach ($this->permissions as $vP)
      if ($vP['to_group_id'] == $gid && $vP['user_recipient_count'] == $this->allowCompleteGroup)
        return true;
    return false;
  }

  public function clearAddressingPermissions () {
    $this->user_group = null;
    $this->permissions = null;
  }



  public function mailAccessPwdTest ($uid, $pwd, &$lang = null) {
    $userRow = $this->db->sql_fetchrow($this->db->sql_query("SELECT user_id, user_lang, mail_access_password FROM ".USERS_TABLE." WHERE (user_type = ".USER_NORMAL." OR user_type = ".USER_FOUNDER.") AND user_id = ".intval($uid)));
    if ($userRow && strlen($userRow['mail_access_password']) > 0 && $userRow['mail_access_password'] == $pwd) {
      if (isset($lang)) $lang = strval($userRow['user_lang']);
      return true;
    } else return false;
  }

  /*
   setSubscription
   no additional permission check!
   $subs: 0:remove subscription  1:set to standard subscription  2:set to enduring subscription  -1:sets enduring subscription (if enduring) to standard
   returns true, if subscription was set successfully
  */
  public function setSubscription ($uid, $tid, $subs) {
    if ($subs < -1 || $subs > 2) return false;
    if ($uid <= 0 || $tid <= 0) return false;
    if ($subs == -1) {
      $this->db->sql_query('UPDATE '.$this->table_prefix.'topics_watch SET notify_status = 1, enduring = 0 WHERE user_id = '.intval($uid).' AND topic_id = '.intval($tid));
    } else {
      $existingRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT enduring FROM '.$this->table_prefix."topics_watch WHERE topic_id = ".intval($tid)." AND user_id = ".intval($uid)));
      if ($existingRow) {
        if ($subs == 0) $this->db->sql_query('DELETE FROM '.$this->table_prefix.'topics_watch WHERE user_id = '.intval($uid).' AND topic_id = '.intval($tid));
        else $this->db->sql_query('UPDATE '.$this->table_prefix.'topics_watch SET enduring = '.(($subs == 2)?1:0).' WHERE user_id = '.intval($uid).' AND topic_id = '.intval($tid));
      } else if ($subs != 0)
        $this->db->sql_query('INSERT INTO '.$this->table_prefix.'topics_watch (topic_id, user_id, notify_status, enduring) VALUES ('.intval($tid).', '.intval($uid).', 0, '.(($subs == 2)?1:0).')');
    }
    return true;
  }




  public function userCanReadForum ($uid, $fid) {
    if (gettype($uid) != 'integer' || gettype($fid) != 'integer') throw 'Type Missmatch!';
    $p = $this->auth->acl_user_raw_data($uid, false, $fid);
    $read = '-'.$p[$uid][$fid]['f_read'];
    if (!$this->groups['u'.$uid])
      $this->groups['u'.$uid] = $this->db->sql_query('SELECT group_id FROM '.$this->table_prefix.'user_group WHERE user_id = '.intval($uid));
    foreach ($this->groups['u'.$uid] as $group) {
      if ($read == '-0') break;
      $p = $this->auth->acl_group_raw_data($group['group_id'], false, $fid);
      if ('-'.$p[$group['group_id']][$fid]['f_read'] != '-') $read = '-'.$p[$group['group_id']][$fid]['f_read'];
    }
    return $read == '-1';
  }

  /*
   groupCanReadForum
   return: 0:group cannot read forum   1:some members of group cannot read forum   2:complete group can read forum
  */
  public function groupCanReadForum ($gid, $fid) {
    if (gettype($gid) != 'integer' || gettype($fid) != 'integer') throw 'Type Missmatch!';
    $p = $this->auth->acl_group_raw_data($gid, false, $fid);
    if ('-'.$p[$gid][$fid]['f_read'] == '-0') return 0;
    $groupusers = $this->db->sql_query('SELECT user_id FROM '.$this->table_prefix.'user_group'.' WHERE group_id = '.$gid);
    foreach ($groupusers as $groupuser)
      if (!$this->userCanReadForum(intval($groupuser['user_id']), $fid))
        return 1;
    return 2;
  }
}
