<?php
/**
*
* @package msta EmailOnTopic Extension
*
*/

namespace msta\emailontopic\includes\bridges;

class MailMessenger extends \msta\emailontopic\includes\BasicBridge {
  const SHORT_NAME = 'email';
  private $gnupg = null;

  public static $removeHeaders = false;
  private static $instance = null;
  public static function getInstance ($phpbb_container, \phpbb\db\driver\driver_interface $db, \phpbb\config\config $config, \phpbb\event\dispatcher_interface $phpbb_dispatcher, $phpbb_root_path) { // \phpbb\language\language $language
    if (!$config['email_enable']) return null;
    if (self::$instance == null)
      self::$instance = new MailMessenger($phpbb_container, $db, $config, $phpbb_dispatcher, $phpbb_root_path);
    return self::$instance;
  }


  public function searchAvaillableReceivers ($name) {
    $db_groups = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'groups'.' WHERE UPPER(group_name) LIKE '."'".$this->db->sql_escape(strtoupper($name))."%'".' LIMIT 5');
    $db_users = $this->db->sql_query('SELECT * FROM '.USERS_TABLE.' WHERE (user_type = '.USER_NORMAL.' OR user_type = '.USER_FOUNDER.') AND UPPER(username) LIKE '."'".$this->db->sql_escape(strtoupper($name))."%'".' LIMIT 5');
    $info = '';
    $n = 0;
    foreach ($db_groups as $group) {
      if ($n >= 5) break;
      if ($info != '') $info .= "\t";
      $info .= '*'.$group['group_name'];
      $n++;
    }
    foreach ($db_users as $user) {
      if ($n >= 5) break;
      if ($info != '') $info .= "\t";
      $info .= $user['username'];
      $n++;
    }
    return new \Symfony\Component\HttpFoundation\Response($info, 200);
  }

  /*
   prepareGnupg
   Initializes this->gnupg, if not already done. Creates a folder with denied public read acces to store the keys, if not already done.
   On success, this->gnupg will be a valid instance, null otherwise.
  */
  private function prepareGnupg () {
    if (intval($this->config['emailontopic_gnupg_enabled']) == 0)
      return;
    if (!$this->gnupg) {
      $path = $this->phpbb_root_path.'files/gnupg/';
      $dirOk = is_dir($path);
      if (!$dirOk) {
        $dirOk = mkdir($path, 0700);
        file_put_contents($path.'.htaccess', 'deny from all');
      }
      if ($dirOk) putenv("GNUPGHOME=".$path);
      else return;
      try {
        if (intval($this->config['emailontopic_gnupg_enabled']) == 2)
          $this->gnupg = new \gnupg();
        else {
          include_once dirname(__FILE__).'/../../vendor/autoload.php';
          require_once dirname(__FILE__).'/../../vendor/singpolyma/openpgp-php/lib/gnupgReplacement.php';
          $this->gnupg = new \gnupgReplacement();
        }
      } catch (Exception $e) {
        $this->gnupg = null;
      }
      if (!$this->gnupg) return;
    }
    $this->gnupg->clearsignkeys();
    $this->gnupg->clearencryptkeys();
    $this->gnupg->cleardecryptkeys();
  }

  /*
   importPgpKey
   Imports user public keys or forum secret key (when $currentUserId is 0). FileName must be the path to a valid .asc-file on the server.
  */
  public function importPgpKey ($fileName, $currentUserId) {
    $key = file_get_contents($fileName);
    $this->prepareGnupg();
    $info = $this->gnupg->import($key);
    if ($info === false) return false;
    $fp = $info['fingerprint'];
    if (!$fp) return false;
    if ($currentUserId == 0) {
      if ($info['secretimported'] == 0 && $info['secretunchanged'] == 0)
        return false;
      $this->config->set('emailontopic_forum_pgp_key', $fp);
    } else {
      if ($info['secretimported'] != 0 || $info['secretunchanged'] != 0)
        die('SECURITY ALERT: You have uploaded a private key! Revoke this key, generate a new one and try again! This private key is now stored on the server. This cannot be undone.');
      $rowExists = $this->db->sql_fetchrow($this->db->sql_query("SELECT user_id FROM ".$this->table_prefix."profile_fields_data WHERE user_id = ".intval($currentUserId)));
      if ($rowExists) $this->db->sql_query("UPDATE ".$this->table_prefix."profile_fields_data SET pf_pgp_fingerprint = '".$this->db->sql_escape($fp)."' WHERE user_id = ".intval($currentUserId));
      else die('Currently there is no row for additional user fields for current user in the database. Try to set any content for user profile fields like interests and try again.');//$this->db->sql_query("INSERT INTO ".$this->table_prefix."profile_fields_data (user_id, pf_pgp_fingerprint) VALUES (".intval($currentUserId).", '".$this->db->sql_escape($fp)."')");
    }
    return true;
  }

  public function getPublicKey () {
    if (!$this->config['emailontopic_forum_pgp_key']) return '';
    $this->prepareGnupg();
    if (!$this->gnupg) return '';
    return $this->gnupg->export($this->config['emailontopic_forum_pgp_key']);
  }

  /*
   encryptOptSign
   Forces encryption with a given fingerprint. If encryption is not possible, result will be empty.
   The signature key fingerprint ist optional. If not given, the message remains unsigned.
  */
  private function encryptOptSign ($encFp, $sgnFp, $message) {
    if (!$this->gnupg || !$encFp) return '';
    $this->gnupg->addencryptkey($encFp);
    if ($sgnFp) {
      $this->gnupg->addsignkey($sgnFp);
      return $this->gnupg->encryptsign($message);
    } else {
      return $this->gnupg->encrypt($message);
    }
  }

  public function sendSubscription($inputBridge, $inputInfo, $targetUser, $template, $message, $subject, $subjectSecure, $reason, $currentUser = null) {
    include_once($this->phpbb_root_path.'includes/functions_messenger.php');
    // init queue if used
    $useQueue = false;
    if ($this->config['email_package_size'] && $reason != 'group_notify' && $reason != 'user_explicit_notify') {
      $useQueue = true;
      if (!$this->queue) {
        $this->queue = new \queue();
        $this->queue->init('email', $this->config['email_package_size']);
      }
    }
    // generate mail content
    $clearMail = false;
    if ($currentUser) {
      $clearMail = intval($currentUser->data['mail_visible_override']) == 1;
      if (!$clearMail)
        $clearMail = intval($this->config['board_hide_emails']) == 0;
      if ($currentUser->data['user_allow_viewemail'] == 0)
        $clearMail = false;
    }
    $encode_eol = ($this->config['smtp_delivery']) ? "\r\n" : PHP_EOL;
    $msgId = md5(unique_id(time()).$targetUser['uid']).'@'.($this->config['server_name'])?:'unknown-server';
    $willBeE2E = ($this->config['emailontopic_gnupg_enabled'] == 1) && $targetUser['pgp'];
    $template->assign_vars(array(
      'msg' => $message,
      'e2e_availlable' => $this->config['emailontopic_gnupg_enabled'],
      'will_be_e2e' => (($willBeE2E)?1:0),
      'formatting_level' => 5,
      'bridge' => self::SHORT_NAME,
    ));
    $subscriptionMessage = trim($template->assign_display('html_post_notification'));
    // encrypt or sign message if possible or required
    $rand = rand(10000000, 99999999);
    if ($willBeE2E) {
      $this->prepareGnupg();
      $headerContentType = 'multipart/encrypted; protocol="application/pgp-encrypted"; boundary="EncryptionPartBoundary'.$rand.'"';
      $subscriptionMessage = ""
       ."This is an OpenPGP/MIME encrypted message (RFC 4880 and 3156)\n"
       ."--EncryptionPartBoundary".$rand."\n"
       ."Content-Type: application/pgp-encrypted\n"
       ."Content-Description: PGP/MIME version identification\n"
       ."\n"
       ."Version: 1\n"
       ."\n"
       ."--EncryptionPartBoundary".$rand."\n"
       .'Content-Type: application/octet-stream; name="encrypted.asc"'."\n"
       ."Content-Description: OpenPGP encrypted message\n"
       .'Content-Disposition: inline; filename="encrypted.asc"'."\n"
       ."\n"
       .$this->encryptOptSign($targetUser['pgp'], $this->config['emailontopic_forum_pgp_key'], ""
        .'Content-Type: multipart/mixed; boundary="ContentPartBoundary'.$rand.'"; protected-headers="v1"'."\n"
        ."Subject: ".mail_encode($subjectSecure)."\n"
        ."From: <".$this->config['board_email'].">\n"
        ."To: <".trim($targetUser['email']).">\n"
        ."Message-ID: <".$msgId.">\n"
        ."\n"
        ."--ContentPartBoundary".$rand."\n"
        ."Content-Type: text/html; charset=utf-8\n"
        ."\n"
        .$subscriptionMessage
        ."\n"
        ."--ContentPartBoundary".$rand."--"
       )
       ."\n"
       ."--EncryptionPartBoundary".$rand."--";
      if (!$this->gnupg) $lastErr = 'GNUPG not initialized!';
      else $lastErr = $this->gnupg->geterror();
      if ($lastErr !== false) {
        $headerContentType = 'text/plain';
        $subscriptionMessage = "An error occured when encrypting this mail: ".$lastErr;
      }
    } else {
      $headerContentType = 'text/html; charset=utf-8';
    }
    // header
    $headers = array(
      'Reply-To: <'.(($clearMail)?$currentUser->data['user_email']:$this->config['board_email']).'>',
      'Return-Path: <'.$this->config['board_email'].'>',
      'Sender: <'.$this->config['board_email'].'>',
      'MIME-Version: 1.0',
      'Message-ID: <'.$msgId.'>',
      'Date: '.date('r', time()),
      'Content-Type: '.$headerContentType,
      'Content-Transfer-Encoding: 8bit',
      'X-Priority: '.MAIL_NORMAL_PRIORITY,
      'X-MSMail-Priority: Normal',                    // (($this->mail_priority == MAIL_LOW_PRIORITY) ? 'Low' : (($this->mail_priority == MAIL_NORMAL_PRIORITY) ? 'Normal' : 'High'))
      'X-Mailer: phpBB3',
      'X-MimeOLE: phpBB3',
      'X-phpBB-Origin: phpbb://'.str_replace(array('http://', 'https://'), array('', ''), generate_board_url()),
    );
    $vars = array('headers');
    extract($this->phpbb_dispatcher->trigger_event('core.modify_email_headers', compact($vars)));
    // mail adresses
    $addresses = array('to' => array(
      array(
        'name' => trim($targetUser['name']),
        'email' => trim($targetUser['email']),
      ),
    ));
    // Send message ...
    if ($useQueue) {
      $this->queue->put('email', array(
        'to'			=> trim($targetUser['email']),
        'addresses'		=> $addresses,
        'subject'		=> $subject,
        'msg'			=> $subscriptionMessage,
        'headers'		=> $headers,
      ));
      return true;
    } else {
      $err_msg = '';
      if ($this->config['smtp_delivery']) {
        return smtpmail($addresses, mail_encode($subject), wordwrap(utf8_wordwrap($subscriptionMessage), 997, "\n", true), $err_msg, $headers);
      } else {
        return phpbb_mail(trim($targetUser['email']), $subject, $subscriptionMessage, $headers, PHP_EOL, $err_msg);
      }
    }
  }

  public function finalSteps () {
    if ($this->queue)
      $this->queue->save();
  }


  private function valOrKey ($arr, $key) {
    if (array_key_exists($key, $arr)) return $arr[$key];
    else return $key;
  }

  /*
   If the who-visited-this-topic extension is installed, add the user with $uid to the topics visitors.
  */
  private function wvttMarkRead($uid, $tid, $mailPostTime) {
    if ($this->config['whovisitedthistopic_allow_topics'] == 1) {
      $wvttRow = $this->db->sql_fetchrow($this->db->sql_query("SELECT counter_user, date FROM ".$this->table_prefix."whovisitedthistopic WHERE user_id = ".intval($uid)." AND topic_id = ".intval($tid)));
      if ($wvttRow) {
        if (intval($wvttRow['date']) < intval($mailPostTime))
          $this->db->sql_query("UPDATE ".$this->table_prefix."whovisitedthistopic SET date = ".intval($mailPostTime).", counter_user = ".(intval($wvttRow['counter_user'])+1)." WHERE user_id = ".intval($uid)." AND topic_id = ".intval($tid));
      } else {
        $this->db->sql_query("INSERT INTO ".$this->table_prefix."whovisitedthistopic (date, counter_user, user_id, topic_id) VALUES (".intval($mailPostTime).", 1, ".intval($uid).", ".intval($tid).")");
      }
    }
  }

  /*
   Generates an image with information on newer posts in a topic. The given ids must be integer values, $pwd is the email access password of the user.
  */
  public function postinfoimg ($uid, $fid, $tid, $pid, $lang, $mailAccessCorrect=false, $userCanReadForum=false) {
    if (gettype($uid) != 'integer' || gettype($fid) != 'integer' || gettype($tid) != 'integer' || gettype($pid) != 'integer') throw 'Type Missmatch!';
    $hasAccess = false;
    $newerPostCount = 0;
    $langPath = '';
    if ($uid > 0 && $fid > 0 && $tid > 0 && $pid > 0) {
      $mailPostTime = 0;
      if ($mailAccessCorrect) {
        $hasAccess = true;
        $langPath = $this->phpbb_root_path.'ext/msta/emailontopic/language/';
        if ($lang != '' && is_dir($langPath.$lang.'/')) $langPath .= $lang;
        else if ($lang != '' && is_dir($langPath.explode('_', $lang)[0].'/')) $langPath .= explode('_', $lang)[0];
        else if (is_dir($langPath.$this->config['default_lang'].'/')) $langPath .= $this->config['default_lang'];
        else if (is_dir($langPath.explode('_', $this->config['default_lang'])[0].'/')) $langPath .= explode('_', $this->config['default_lang'])[0];
        else $langPath = '';
      }
      if ($hasAccess && !$userCanReadForum) $hasAccess = false;
      if ($hasAccess) {
        $mailPostRow = $this->db->sql_fetchrow($this->db->sql_query("SELECT post_time, topic_id, forum_id FROM ".$this->table_prefix."posts WHERE post_id = ".$pid));
        if ($mailPostRow && $mailPostRow['topic_id'] == $tid && $mailPostRow['forum_id'] == $fid) $mailPostTime = intval($mailPostRow['post_time']);
        else $hasAccess = false;
      }
      if ($hasAccess) {
        $postCountRow = $this->db->sql_fetchrow($this->db->sql_query("SELECT COUNT(*) AS 'count' FROM ".$this->table_prefix."posts WHERE topic_id = ".$tid." AND post_time > ".$mailPostTime));
        if (!$postCountRow) $hasAccess = false;
        else {
          $newerPostCount = intval($postCountRow['count']);
          $this->wvttMarkRead($uid, $tid, $mailPostTime);
        }
      }
    }
    if ($langPath == '') $langPath = $this->phpbb_root_path.'ext/msta/emailontopic/language/en';
    $lang = array();
    include $langPath.'/main_interface.php';
    if ($hasAccess) {
      if ($newerPostCount == 0) {
        $txt = $this->valOrKey($lang, 'EMAILONTOPIC_MAILINFO_NEWEST');
        $bg = 'Newest';
      } else {
        $txt = $this->valOrKey($lang, 'EMAILONTOPIC_MAILINFO_OUTDATED');
        $txt = join($newerPostCount, explode('%NEWER_POST_COUNT', $txt));
        $bg = 'Outdated';
      }
    } else {
      $txt = $this->valOrKey($lang, 'EMAILONTOPIC_MAILINFO_ERROR');
      $bg = 'Error';
    }
    $image = imagecreatefrompng($this->phpbb_root_path.'ext/msta/emailontopic/styles/emailinfo/EmailHintBackground'.$bg.'.png');
    $textcolor = imagecolorallocate($image, 0, 0, 0);
    imagettftext($image, 9, 0, 5, 15, $textcolor, $this->phpbb_root_path.'ext/msta/emailontopic/styles/emailinfo/font.ttf', $txt);
    imagestring($image, 1, 5, 40, date("Y.m.d H:i"), $textcolor);
    header('Cache-Control: no-cache');
    header('Content-Type: image/png');
    header('Content-Disposition: inline; filename="postinfo.png"');
    imagepng($image);
  }
}
