<?php
/**
*
* @package msta EmailOnTopic Extension
*
*/

namespace msta\emailontopic\includes\bridges;

class TelegramBridge extends \msta\emailontopic\includes\BasicBridge {
  const SHORT_NAME = 'telegram';
  private $lastElementIsBlock = false;
  private static $instance = null;
  public static function getInstance ($phpbb_container, \phpbb\db\driver\driver_interface $db, \phpbb\config\config $config, \phpbb\event\dispatcher_interface $phpbb_dispatcher, $phpbb_root_path) {
    if (!$config['telegram_webhook_pass']) return null;
    if (self::$instance == null)
      self::$instance = new TelegramBridge($phpbb_container, $db, $config, $phpbb_dispatcher, $phpbb_root_path);
    return self::$instance;
  }

  public static function linkedChannelExists ($topicId, $db, $tablePrefix) {
    return $db->sql_fetchrow($db->sql_query('SELECT COUNT(*) as cnt FROM '.$tablePrefix.'telegram_links WHERE forum_link_id = '.intval($topicId).' AND (link_type = 2 OR link_type = 3) AND active = 1'))['cnt'] > 0;
  }

  /*
   jsonVal
   returns the value of a (sub) object of the given $json array.
   $key is an array representing the tree keys to the property.
   If $dev is not null and the datatype of the json does not match the datatype of $dev or the tree does not exist, $dev is returned.
   If $ucrEscapeStrings is true and the result is a string, 4-byte-unicode-chars will be html-escaped for mysql.
  */
  private function jsonVal ($json, $key, $def=null, $ucrEscapeStrings=true) {
    $o = $json;
    for ($i = 0; $i < count($key); $i++) {
      if (!is_array($o)) {
        $o = null;
        break;
      }
      if (array_key_exists($key[$i], $o)) $o = $o[$key[$i]];
      else {
        $o = null;
        break;
      }
    }
    if ($ucrEscapeStrings && gettype($o) == 'string') $o = utf8_encode_ucr($o);
    if ($def === null || gettype($o) == gettype($def)) return $o;
    else if (gettype($o) == 'integer' && gettype($def) == 'string') return strval($o); // this conversation is save and always works
    else return $def;
  }

  private function sqlInsertStr ($obj) {
    $keyStr = '';
    $valStr = '';
    foreach ($obj as $k => $v) {
      if ($keyStr != '') {
        $keyStr .= ', ';
        $valStr .= ', ';
      }
      $keyStr .= $k;
      if (gettype($v) == 'integer') $valStr .= $v;
      else if (gettype($v) == 'string') $valStr .= "'".$this->db->sql_escape($v)."'";
      else die('Unsupported variable type for SQL!');
    }
    return '('.$keyStr.') VALUES ('.$valStr.')';
  }

  private function sqlUpdateStr ($obj, $keys) { // creates an sql-update-sequence for all given $keys of $obj
    $uStr = '';
    foreach ($keys as $k) {
      if ($uStr != '') $uStr .= ', ';
      $uStr .= $k.' = ';
      if (gettype($obj[$k]) == 'integer') $uStr .= $obj[$k];
      else if (gettype($obj[$k]) == 'string') $uStr .= "'".$this->db->sql_escape($obj[$k])."'";
      else die('Unsupported variable type for SQL!');
    }
    return 'SET '.$uStr;
  }

  /*
   formatHTMLForTelegram
   Converts the given $node and it's subtree into a HTML-string and only contains HTML-tags, that are supported by telegram.
  */
  private function formatHTMLForTelegram ($node, $parentTree) {
    $thisElementIsBlock = false;
    if ($node->nodeType == 3) {
      $s = htmlspecialchars(trim($node->nodeValue));
      if ($s == '' || $node->isElementContentWhitespace()) {
        return '';
      } else {
        if ($this->lastElementIsBlock) {
          $this->lastElementIsBlock = false;
          return "\n".$s;
        } else return $s;
      }
    } else if ($node->nodeType == 1) {
      if ($top) $n = '';
      else $n = $node->nodeName;
      if ($n == 'strong') $n = 'b';
      if ($n == 'h1' || $n == 'h2' || $n == 'h3' || $n == 'h4') {
        $n = 'b';
        $thisElementIsBlock = true;
      }
      if ($n == 'div' || $n == 'p') $thisElementIsBlock = true;
      if ($n == 'em') $n = 'i';
      if ($n == 'span' && $node->getAttribute('class') == 'abbc3_strike') $n = 's';
      if ($n == 'br') return "\n";
      $tagAllowed =
        ($n == 'a' && $this->treeContainsOnly($parentTree, array('b', 'i', 'u', 's')) && (substr($node->getAttribute('href'), 0, 7) == 'http://' || substr($node->getAttribute('href'), 0, 8) == 'https://')) ||
        ($n == 'b' && $this->treeContainsOnly($parentTree, array('a', 'i', 'u', 's'))) ||
        ($n == 'i' && $this->treeContainsOnly($parentTree, array('a', 'b', 'u', 's'))) ||
        ($n == 'u' && $this->treeContainsOnly($parentTree, array('a', 'i', 'b'))) ||
        ($n == 's' && $this->treeContainsOnly($parentTree, array('a', 'i', 'b'))) ||
        ($n == 'code' && $this->treeContainsOnly($parentTree, array())) ||
        ($n == 'pre' && $this->treeContainsOnly($parentTree, array('code')));
      if ($this->lastElementIsBlock || $thisElementIsBlock) $txt = "\n";
      else $txt = '';
      $this->lastElementIsBlock = false;
      if ($tagAllowed) {
        if ($n == 'a') $txt .= '<a href="'.$node->getAttribute('href').'">';
        else $txt .= '<'.$n.'>';
        $parentTree[] = $n;
      }
      $childNodes = $node->childNodes;
      for ($i = 0; $i < $childNodes->length; $i++)
        $txt .= $this->formatHTMLForTelegram($childNodes[$i], $parentTree);
      if ($tagAllowed)
        $txt .= '</'.$n.'>';
      $this->lastElementIsBlock = $thisElementIsBlock;
      return $txt;
    } else return '';
  }

  private function compareEntiriesByStartAndLength ($a, $b) { // required by teleramEntitiesDecode
    if ($a['start'] > $b['start']) return 1;
    if ($a['start'] < $b['start']) return -1;
    if ($a['length'] < $b['length']) return 1;
    if ($a['length'] > $b['length']) return -1;
    return 0;
  }

  private function entityToTag ($e, $opening) { // required by teleramEntitiesDecode
    $tag = '';
    $slash = ($opening)?'':'/';
    if (array_key_exists('type', $e))
      switch (strtolower(strval($e['type']))) {
        case 'text_link':
          $tag = '<'.$slash.'a';
          if ($opening && array_key_exists('url', $e))
            $tag .= ' href="'.utf8_encode_ucr(htmlspecialchars(strval($e['url']))).'" target="_blank"';
          $tag .= '>';
          break;
        case 'url':
          $tag = '<'.$slash.'a';
          if ($opening)
            $tag .= ' href="javascript:" onclick="this.href = this.textContent; this.target = '."'_blank'".';"';
          $tag .= '>';
          break;
        case 'bold':
          $tag = '<'.$slash.'b>';
          break;
        case 'italic':
          $tag = '<'.$slash.'i>';
          break;
        case 'underline':
          $tag = '<'.$slash.'u>';
          break;
        case 'strikethrough':
          $tag = '<'.$slash.'s>';
          break;
        case 'code':
        case 'pre':
          $tag = '<'.$slash.'pre';
          if ($opening && array_key_exists('language', $e))
            $tag .= ' language="'.utf8_encode_ucr(htmlspecialchars(strval($e['language']))).'"';
          $tag .= '>';
          break;
        case 'text_mention':
          $tag = '<'.$slash.'span';
          if ($opening && array_key_exists('user', $e) && is_array($e['user'])) {
            $userName = ((array_key_exists('first_name', $e['user']))?utf8_encode_ucr(htmlspecialchars(strval($e['user']['first_name']))).' ':'').((array_key_exists('last_name', $e['user']))?utf8_encode_ucr(htmlspecialchars(strval($e['user']['last_name']))):'');
            if ($userName != '')
              $tag .= ' onclick="window.location.href = '."'memberlist.php?first_char=".rawurlencode(strtolower($userName))."';".'" style="cursor:pointer; color:gray;" title="'.$userName.'"';
          }
          $tag .= '>';
          break;
        case 'email':
          $tag = '<'.$slash.'a ';
          if ($opening)
            $tag .= ' href="javascript:" onclick="this.href = '."'mailto:'".'+this.textContent; "';
          $tag .= '>';
          break;
        case 'phone_number':
          $tag = '<'.$slash.'a ';
          if ($opening)
            $tag .= ' href="javascript:" onclick="this.href = '."'callto:'".'+this.textContent; "';
          $tag .= '>';
          break;
        case 'hashtag':
          $tag = '<'.$slash.'a ';
          if ($opening)
            $tag .= ' href="javascript:" onclick="this.href = '."'search.php?keywords='+encodeURIComponent(this.textContent.split('#').join(''));".'"';
          $tag .= '>';
          break;
        case 'mention':
          $tag = '<'.$slash.'a ';
          if ($opening)
            $tag .= ' href="javascript:" onclick="this.href = '."'memberlist.php?first_char='+encodeURIComponent(this.textContent.split('@').join(''));".'"'; // TODO: if a telegram-user-id :: forum-user-id map exists, add better linkage here
          $tag .= '>';
          break;
      }
    return $tag;
  }

  /*
   teleramEntitiesDecode
   converts a telegram unicode string with extra formatting entities into a forum-valid html-representation.
   4-byte-unicode-chars will be escaped for mysql. Html-Special-Chars will be escaped in $originalText.
  */
  private function teleramEntitiesDecode ($originalText, $entities=null) {
    if (!$entities)
      return join('<br/>', explode("\n", utf8_encode_ucr(htmlspecialchars($originalText))));
    // convert original text to uft 16 array and escape the chars
    $utf16Text = preg_split('//u', $originalText, -1, PREG_SPLIT_NO_EMPTY);
    $i = 0;
    while ($i < count($utf16Text)) {
      switch ($utf16Text[$i]) {
        case "\n":
          $utf16Text[$i] = '<br>';
          break;
        case "&":
          $utf16Text[$i] = '&amp;';
          break;
        case '"':
          $utf16Text[$i] = '&quot;';
          break;
        case "'":
          $utf16Text[$i] = '&#039;';
          break;
        case "<":
          $utf16Text[$i] = '&lt;';
          break;
        case ">":
          $utf16Text[$i] = '&gt;';
          break;
        default:
          if (strlen($utf16Text[$i]) > 2) {
            for ($j = 2; $j < strlen($utf16Text[$i]); $j += 2)
              array_splice($utf16Text, $i+1, 0, array('')); // telegram counts the entity-positions as utf-16 code points...
            $utf16Text[$i] = utf8_encode_ucr($utf16Text[$i]);
          }
      }
      $i++;
    }
    // ensure correctness of all entities
    for ($i = 0; $i < count($entities); $i++) {
      if (array_key_exists('offset', $entities[$i])) $entities[$i]['start'] = intval($entities[$i]['offset']);
      else $entities[$i]['start'] = 0;
      if ($entities[$i]['start'] < 0) $entities[$i]['start'] = 0;
      if ($entities[$i]['start'] > count($utf16Text)) $entities[$i]['start'] = count($utf16Text);
      if (array_key_exists('length', $entities[$i])) $entities[$i]['stop'] = intval($entities[$i]['length']) + $entities[$i]['start'];
      else $entities[$i]['stop'] = $entities[$i]['start'];
      if ($entities[$i]['stop'] < $entities[$i]['start']) $entities[$i]['stop'] = $entities[$i]['start'];
      if ($entities[$i]['stop'] > count($utf16Text)) $entities[$i]['stop'] = count($utf16Text);
      $entities[$i]['length'] = $entities[$i]['stop'] - $entities[$i]['start'];
      unset($entities[$i]['offset']);
    }
    // split and sort
    $i = 0;
    while ($i < count($entities)) {
      // make entities[i] the one with the largest length of all entities with index >= i
      $maxI = $i;
      $j = count($entities)-1;
      while ($j > $i) {
        if ($entities[$j]['length'] > $entities[$maxI]['length']) $maxI = $j;
        $j--;
      }
      $copy = $entities[$maxI];
      $entities[$maxI] = $entities[$i];
      $entities[$i] = $copy;
      // split later entities at the borders of [i]
      $j = count($entities)-1;
      while ($j > $i) {
        if ($entities[$j]['start'] < $entities[$i]['start'] && $entities[$j]['stop'] > $entities[$i]['start']) {
          $newE = $entities[$j]; // clones because it's array
          $entities[$j]['stop'] = $entities[$i]['start'];
          $entities[$j]['length'] = $entities[$j]['stop'] - $entities[$j]['start'];
          $newE['start'] = $entities[$i]['start'];
          $newE['length'] = $newE['stop'] - $newE['start'];
          $entities[] = $newE;
        } else if ($entities[$j]['start'] < $entities[$i]['stop'] && $entities[$j]['stop'] > $entities[$i]['stop']) {
          $newE = $entities[$j]; // clones because it's array
          $entities[$j]['stop'] = $entities[$i]['stop'];
          $entities[$j]['length'] = $entities[$j]['stop'] - $entities[$j]['start'];
          $newE['start'] = $entities[$i]['stop'];
          $newE['length'] = $newE['stop'] - $newE['start'];
          $entities[] = $newE;
        }
        $j--;
      }
      $i++;
    }
    // sort entities by tree structure
    usort($entities, array($this, "compareEntiriesByStartAndLength"));
    // go through tree and create output
    $formattedText = '';
    $originalPos = 0;
    $tree = array(['start' => -1, 'stop' => count($utf16Text)+1, 'length' => count($utf16Text)+1]); // virtual root entity, truely surrounding the complete array
    $i = 0;
    while (count($tree) > 0) { // entities[i] is the next opening entity in list
      if ($i < count($entities) && $tree[count($tree)-1]['stop'] > $entities[$i]['start']) { // push tree
        $tree[] = $entities[$i];
        $formattedText .= join('', array_slice($utf16Text, $originalPos, $entities[$i]['start']-$originalPos));
        $originalPos = $entities[$i]['start'];
        $formattedText .= $this->entityToTag($entities[$i], true);
        $i++;
      } else { // pop tree
        $p = array_pop($tree);
        if (count($tree) == 0)
          $formattedText .= join('', array_slice($utf16Text, $originalPos));
        else {
          $formattedText .= join('', array_slice($utf16Text, $originalPos, $p['stop']-$originalPos));
          $originalPos = $p['stop'];
          $formattedText .= $this->entityToTag($p, false);
        }
      }
    }
    return $formattedText;
  }

  /*
   newPostFromTelegram
   $text must be the filally html-escaped text for the forum; use teleramEntitiesDecode on the raw json result
   return value: array of [errorString, newPostId, topicId, forumId]
  */
  private function newPostFromTelegram ($telegram_chat_id, $msgId, $telegramUser, $telegramUserId, $editDate, $date, $text, $subject='#telegram') {
    if (gettype($telegram_chat_id) != 'integer' || gettype($msgId) != 'integer' || gettype($telegramUser) != 'string' || gettype($telegramUserId) != 'integer' || gettype($editDate) != 'integer' || gettype($date) != 'integer' || gettype($text) != 'string' || gettype($subject) != 'string') throw 'Type Missmatch!';
    $newPostId = 0;
    $linkRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_links WHERE telegram_chat_id = '.$telegram_chat_id));
    if (!$linkRow || !($linkRow['link_type'] == 1 || $linkRow['link_type'] == 3))
      return ['No topic is linked with this chat id. Ignoring post...', 0, 0, 0];
    $topicRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT topic_id, forum_id, topic_delete_time FROM '.$this->table_prefix.'topics WHERE topic_id = '.intval($linkRow['forum_link_id'])));
    if (!$topicRow || $topicRow['topic_delete_time'] != 0)
      return ['The topic linked with this chat id does not (longer) exist. Ignoring post...', 0, 0, 0];
    $info = array(
      'topic_id' => intval($topicRow['topic_id']),
      'forum_id' => intval($topicRow['forum_id']),
      'poster_id' => $telegramUserId,
      'post_time' => $date,
      'post_edit_time' => $editDate,
      'post_visibility' => 1,
      'post_subject' => utf8_encode_ucr(htmlspecialchars($subject)),
      'post_text' => '<telegramuser>'.htmlspecialchars($telegramUser).'</telegramuser><t>'.$text.'</t>',
      'post_checksum' => '',
      'enable_bbcode' => 0,
    );
    $info['post_checksum'] = md5($info['post_text']);
    if ($info['post_edit_time'] == 0) {
      $lastId = 0;
      $existingMessageTest = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id FROM '.$this->table_prefix."posts WHERE post_checksum = '".$this->db->sql_escape($info['post_checksum'])."' AND post_time = ".$date));
      if (!$existingMessageTest)
        $existingMessageTest = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id FROM '.$this->table_prefix."telegram_messages WHERE telegram_chat_id = ".$telegram_chat_id." AND telegram_message_id = ".$msgId));
      if ($existingMessageTest)
        return ['Message already send; ignoring double post...', 0, 0, 0]; // for whatever reason, the message could already exist and must not be re-posted. Maybe if an error occured and telegram resend the message.
      else {
        if ($this->db->sql_query('INSERT INTO '.$this->table_prefix.'posts '.$this->sqlInsertStr($info)))
          $newPostId = intval($this->db->sql_nextid());
        if ($newPostId == 0)
          return ['Error creating post!', 0, 0, 0];
        $this->db->sql_query('INSERT INTO '.$this->table_prefix.'telegram_messages (telegram_chat_id, telegram_message_id, post_id) VALUES ('.$telegram_chat_id.', '.$msgId.', '.$newPostId.')');
        $this->db->sql_query('UPDATE '.$this->table_prefix.'topics SET topic_last_post_id = '.$newPostId.', topic_last_poster_id = '.$telegramUserId.", topic_last_poster_name = 'telegram', topic_last_poster_colour = '', topic_last_post_subject = '#telegram', topic_last_post_time = ".intval($info['post_time']).', topic_posts_approved = topic_posts_approved + 1 WHERE topic_id = '.$info['topic_id']);
        $this->db->sql_query('UPDATE '.$this->table_prefix.'users SET user_posts = user_posts + 1 WHERE user_id = '.$telegramUserId);
        $dispatcher = $this->phpbb_container->get('dispatcher');
        $dispatcher->trigger_event('msta.emailontopic.input_bridge_post_end', array(
          'input' => self::SHORT_NAME,
          'inputInfo' => strval($telegram_chat_id),
          'mode' => 'reply',
          'postId' => $newPostId,
          'author' => 'telegram:'.$telegramUser,
        ));
        return ['Post created.', $newPostId, $info['topic_id'], $info['forum_id']];
      }
    } else {
      $postRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_messages WHERE telegram_chat_id = '.$telegram_chat_id.' AND telegram_message_id = '.$msgId));
      if (!$postRow)
        return ['There is no existing post in forum that corresponds to this message id. Ignoring upate...', 0, 0, 0];
      $postExists = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id, topic_id, forum_id FROM '.$this->table_prefix.'posts WHERE post_id = '.intval($postRow['post_id'])));
      if (!$postExists)
        return ['The post corresponding to this message id was removed from forum. Ignoring upate...', 0, 0, 0];
      $this->db->sql_query('UPDATE '.$this->table_prefix.'posts '.$this->sqlUpdateStr($info, array('post_edit_time', 'post_text', 'post_checksum')).', post_edit_count = post_edit_count + 1 WHERE post_id = '.intval($postRow['post_id']));
      $dispatcher->trigger_event('msta.emailontopic.input_bridge_post_end', array(
        'input' => self::SHORT_NAME,
        'inputInfo' => strval($telegram_chat_id),
        'mode' => 'edit',
        'postId' => intval($postExists['post_id']),
        'author' => 'telegram:'.$telegramUser,
      ));
      return ['Post updated.', intval($postExists['post_id']), intval($postExists['topic_id']), intval($postExists['forum_id'])];
    }
    return ['Unknown error!', 0, 0, 0];
  }

  /*
   addAttachmentToPost
   Downloads a file from a given $url to the local server and adds it as an attachment to the post with the given $postId. This post must exist.
   The attachment's $comment is optional and HTML-Code will not be escaped. You have to apply htmlspecialchars() on forign/user-input content given as comment!
   If $dispatchEvent is true, the core.modify_attachment_sql_ary_on_submit will be triggered to allow other extensions to work on the attachment.
   If $topicId, $forumId AND the poster's $userId (which is not always the current user's ID) of the given post are known, they can be given to save one SQL-query.
   $realFileName can be given; if empty, the file name is taken from the url.
   Returns an array holding the error information. If successfull, first array element is 'SUCCESS', second element the ID of the new attachment, and third element again the post ID.
  */
  private function addAttachmentToPost ($url, $postId, $comment='', $dispatchEvent=true, $topicId=0, $forumId=0, $userId=0, $realFileName='') {
    if (getType($url) != 'string' || getType($postId) != 'integer' || getType($comment) != 'string' || getType($dispatchEvent) != 'boolean' || getType($topicId) != 'integer' || getType($forumId) != 'integer' || getType($userId) != 'integer') throw 'Type Missmatch';
    if ($postId <= 0) throw 'Post ID cannot be zero!';
    // if not given, get missing IDs
    if ($topicId == 0 || $forumId == 0 || $userId == 0) {
      $idRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id, topic_id, forum_id, poster_id FROM '.$this->table_prefix.'posts WHERE post_id = '.$postId));
      if (!$idRow)
        return ['POST_NOT_EXISTS'];
      $topicId = intval($idRow['topic_id']);
      $forumId = intval($idRow['topic_id']);
      $userId = intval($idRow['poster_id']);
    }
    // check attachment count
    if (intval($this->config['max_attachments']) >= 0) {
      $cnt = $this->db->sql_fetchrow($this->db->sql_query('SELECT COUNT(*) as cnt FROM '.$this->table_prefix.'attachments WHERE post_msg_id = '.$postId))['cnt'];
      if ($cnt >= intval($this->config['max_attachments']))
        return ['MAX_ATTACHMENTS_PER_POST_EXCEEDED'];
    }
    // get required classes
    $upload = $this->phpbb_container->get('files.upload');
    $cache = $this->phpbb_container->get('cache');
    $attach = $this->phpbb_container->get('attachment.upload');
    // load file from remote location to local server
    $upload->set_disallowed_content([]);
    $extensions = $cache->obtain_attach_extensions($forumId);
    $extensionArr = array_keys($extensions['_allowed_']);
    $upload->set_allowed_extensions($extensionArr);
    $tempFile = $upload->handle_upload('files.types.remote', $url);
    if (count($tempFile->error) > 0) {
      $ext = '.';
      if (strrpos($url, '.') !== false) $ext = substr($url, strrpos($url, '.')+1);
      if ($tempFile->error[0] == 'URL_INVALID' && !in_array($ext, $extensionArr)) return ['FILE_EXTENSION_NOT_ALLOWED', 'EXTENSION=.'.htmlspecialchars($ext)];
      return $tempFile->error;
    }
    $tempFileData = [
      'realname' => (($realFileName == '')?$tempFile->get('realname'):htmlspecialchars($realFileName)),
      'size' => $tempFile->get('filesize'),
      'type' => $tempFile->get('mimetype'),
    ];
    // create attachment from temp file
    if (!function_exists('create_thumbnail'))
      require($this->phpbb_root_path.'includes/functions_posting.php');
    $attachmentFileData = $attach->upload('', $forumId, true, $tempFile->get('filename'), false, $tempFileData);
    if (!$attachmentFileData['post_attach'])
      return ['FILE_ATTACH_ERROR'];
    if (count($attachmentFileData['error']) > 0)
      return $attachmentFileData['error'];
    $sql_ary = array(
      'physical_filename'	=> $attachmentFileData['physical_filename'],
      'attach_comment'		=> $comment,
      'real_filename'		=> $attachmentFileData['real_filename'],
      'extension'			=> $attachmentFileData['extension'],
      'mimetype'			=> $attachmentFileData['mimetype'],
      'filesize'			=> $attachmentFileData['filesize'],
      'filetime'			=> $attachmentFileData['filetime'],
      'thumbnail'			=> $attachmentFileData['thumbnail'],
      'is_orphan'			=> 0,
      'in_message'			=> 0,
      'poster_id'			=> $userId,
      'post_msg_id'			=> $postId,
      'topic_id'			=> $topicId,
    );
    if ($dispatchEvent) {
      $dispatcher = $this->phpbb_container->get('dispatcher');
      $vars = array('sql_ary');
      extract($dispatcher->trigger_event('core.modify_attachment_sql_ary_on_submit', compact($vars)));
    }
    $this->db->sql_query('INSERT INTO '.$this->table_prefix.'attachments '.$this->db->sql_build_array('INSERT', $sql_ary));
    $newAttachmentID = intval($this->db->sql_nextid());
    if ($newAttachmentID == 0) return ['SQL_ATTACHMENT_INSERT_ERROR'];
    $this->db->sql_query('UPDATE '.$this->table_prefix.'posts SET post_attachment = 1 WHERE post_id = '.$postId);
    return ['SUCCESS', $newAttachmentID, $postId];
  }

  private function telegramDownloadLink ($fileId) {
    $json = $this->telegramOp('getFile?file_id='.rawurlencode($fileId));
    if ($json) {
      $path = $this->jsonVal($json, array('result', 'file_path'), '');
      if ($path == '')
        return ['', htmlspecialchars($this->jsonVal($json, array('description'), 'TELEGRAM RETURNED NO DOWNLOAD PATH AND NO ERROR INFORMATION'))];
      else
         return ['https://api.telegram.org/file/bot'.$this->config['telegram_bot_key'].'/'.$path, ''];
    } else return ['', 'TELEGRAM RETURNED NO VALID JSON FOR DOWNLOAD PATH'];
  }

  /*
   telegramFileImport
   imports a file from telegram as an attachment of a post. Creates a new post with content $comment for this or uses an existing post, if mediaGroup of the message belongs to an existing post.
   $json is the parsed json object of the telegram query and should contain a message-object.
   $fileId is the telegram file id of the attachment
   $telegram_chat_id, $telegramUserId and $date are the values as 
   $comment must be the finally html-escaped representation for the forum.
  */
  private function telegramFileImport ($telegram_chat_id, $msgId, $telegramUser, $telegramUserId, $editDate, $date, $mediaGroup, $fileId, $comment='', $realFileName='') {
    if (gettype($telegram_chat_id) != 'integer' || gettype($msgId) != 'integer' || gettype($telegramUser) != 'string' || gettype($telegramUserId) != 'integer' || gettype($editDate) != 'integer' || gettype($date) != 'integer' || gettype($mediaGroup) != 'string' || gettype($fileId) != 'string' || gettype($comment) != 'string' || gettype($realFileName) != 'string') throw 'Type Missmatch!';
    // find or create post for media(group)
    $mediaGroupPost = null;
    if ($mediaGroup != '') {
      $mediaGroup = '<!-- '.htmlspecialchars($mediaGroup).' -->';
      $mediaGroupPost = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_msg_id FROM '.$this->table_prefix."attachments WHERE attach_comment = '".$this->db->sql_escape($mediaGroup)."' LIMIT 1"));
    }
    if ($mediaGroupPost)
      $postInfo = ['', intval($mediaGroupPost['post_msg_id']), 0, 0];
    else
      $postInfo = $this->newPostFromTelegram(
        $telegram_chat_id,
        $msgId,
        $telegramUser,
        $telegramUserId,
        $editDate,
        $date,
        $comment
      );
    if ($postInfo[1] == 0)
      return ['NO_POST_FOR_FILE', $postInfo[0]];
    // if this is an update, remove previous attachment from post
    if ($postInfo[0] == 'Post updated.') {
      $attachRows = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'attachments WHERE post_msg_id = '.$postInfo[1]);
      while ($attachRow = $this->db->sql_fetchrow($attachRows)) {
        unlink($this->config['upload_path'].'/'.$attachRow['physical_filename']);
        if (intval($attachRow['thumbnail']) == 1)
          unlink($this->config['upload_path'].'/thumb_'.$attachRow['physical_filename']);
      }
      $this->db->sql_query('DELETE FROM '.$this->table_prefix.'attachments WHERE post_msg_id = '.$postInfo[1]);
    }
    // create html-node for download from telegram; this will be inserted into the post before file download as this can be skipped by php-execution-timeout
    $downloadLinkBegin = '<telegramdownload fileid="'.utf8_encode_ucr(htmlspecialchars($fileId)).'">';
    $downloadLink = $downloadLinkBegin.'<br><a href="/app.php/emailontopic/getinfo?action=telegramdownload&fileid='.rawurlencode($fileId).'" target="_blank" style="display:inline-block; background-color:rgba(200, 200, 255, 0.3); min-width:100px; text-align:center; border-radius:3px; border:1px dotted gray; padding:3px; ">'.(($realFileName == '')?'download':utf8_encode_ucr(htmlspecialchars($realFileName))).'</a></telegramdownload>';
    $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = CONCAT(post_text, '".$this->db->sql_escape($downloadLink)."') WHERE post_id = ".intval($postInfo[1]));
    // get file
    $url = $this->telegramDownloadLink($fileId);
    if ($url[0] == '')
      $err = [$url[1]];
    else
      $err = $this->addAttachmentToPost($url[0], $postInfo[1], $mediaGroup, true, $postInfo[2], $postInfo[3], $telegramUserId, $realFileName);
    // remove html-node for download from telegram, if attachment was created successfully or add error information
    if (count($err) == 0) $err = ['UNKNOWN_ERROR'];
    $postText = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_text FROM '.$this->table_prefix."posts WHERE post_id = ".intval($postInfo[1])));
    if (!$postText) return $err;
    $postText = strval($postText['post_text']);
    $i1 = strpos($postText, $downloadLinkBegin);
    if ($i1 === false) return $err;
    $i2 = strpos($postText, '</telegramdownload>', $i1);
    if ($i2 === false) return $err;
    if ($err[0] == 'SUCCESS')
      $postText = substr($postText, 0, $i1).substr($postText, $i2+19); // just remove extra link, if import was successfull
    else
      $postText = substr($postText, 0, $i2).'<span style="color:red; margin-left:3px; margin-right:15px; ">'.htmlspecialchars(join(', ', $err)).'</span>'.substr($postText, $i2); // otherwise add error information to download link
    $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = '".$this->db->sql_escape($postText)."' WHERE post_id = ".intval($postInfo[1]));
    return $err;
  }

  /*
   attachmentHideScript
   Returns a javascript, that can be placed behind a html-element, that represents an embedded attachment in a post. The script then hides the attachment download field, created by phpBB.
   $embeddedElType is the tag name of the embedded html element.
   $id the attachment id of the corresponding file.
  */
  private function attachmentHideScript ($embeddedElType, $id) {
    return 
      '<script type="text/javascript">'.
        '(function () {'.
          'var vid = document.getElementsByTagName("'.$embeddedElType.'")[document.getElementsByTagName("'.$embeddedElType.'").length-1];'.
          'document.addEventListener("DOMContentLoaded", function() {'.
            'var a = vid.parentElement.parentElement.getElementsByTagName("a");'.
            'var search = "download/file.php?id='.intval($id).'";'.
            'for (var i = a.length-1; i >= 0; i--) '.
              'if (a[i].href.indexOf(search) == a[i].href.length - search.length) {'.
                'var dl = a[i].parentElement.parentElement;'.
                'if (dl.tagName.toLowerCase() != "dl" || dl.className.indexOf("file") < 0) break;'.
                'var dd = dl.parentElement;'.
                'dd.removeChild(dl);'.
                'if (dd.getElementsByTagName("dl").length == 0) '.
                  'dd.parentElement.parentElement.removeChild(dd.parentElement);'.
                'break;'.
              '}'.
          '});'.
        '})();'.
      '</script>';
  }

  /*
   The webhook was called (Telegram->phpBB message). Verify password and evaluate given data.
   $postContent: The received, unconverted file content of the post request to the webhook
   $pwd: The Webhook access password from query string
   $newPostId: If telegram commands to create a new post, this is the id if this new post
   $telegramUser: If given, this is the Username of the telegram user, who posted the message (With html special chars escaped)
  */
  public function telegramInput ($postContent, $pwd, &$newPostId, &$telegramUser) {
    // authenification and json decode
    $resultMessage = 'Received.';
    if ($this->config['telegram_webhook_pass'] == '') return new \Symfony\Component\HttpFoundation\Response('Webhook deactivated!', 403);
    if ($pwd != $this->config['telegram_webhook_pass']) return new \Symfony\Component\HttpFoundation\Response('Webhook password wrong!', 403);
    try { $json = json_decode($postContent, true); }
    catch (Exception $e) { $json = null; }
    if (!$json) return new \Symfony\Component\HttpFoundation\Response('Invalid JSON given!', 400);
    $botId = intval($this->config['telegram_bot_id']);
    if ($botId == 0) return new \Symfony\Component\HttpFoundation\Response('Telegram bot not correctly initialized!', 500);

    // collect some often required values
    if (!array_key_exists('message', $json) && array_key_exists('edited_message', $json))
      $json['message'] = $json['edited_message'];
    $telegram_chat_id = intval($this->jsonVal($json, array('message', 'chat', 'id'), 0));
    $msgId = intval($this->jsonVal($json, array('message', 'message_id'), 0));
    $telegramUser = strval($this->jsonVal($json, array('message', 'from', 'first_name'), ''));
    $telegramUserId = intval($this->config['telegram_user_id']);
    $date = intval($this->jsonVal($json, array('message', 'date'), 0));
    $editDate = intval($this->jsonVal($json, array('message', 'edit_date'), 0));
    $fromId = intval($this->jsonVal($json, array('message', 'from', 'id'), 0));
    $newPostId = -1; // the id of the new post in forum  -1 means, no new post was required, other negative values are special error codes

    foreach ($this->jsonVal($json, array('message', 'new_chat_members'), array()) as $newMember) { // check if this bot is added to a new channel
      if ($this->jsonVal($newMember, array('id'), 0) == $botId) {
        $info = array(
          'telegram_chat_id' => $telegram_chat_id,
          'forum_link_id' => 0,
          'link_type' => 0,
          'chat_name' => substr(htmlspecialchars(strval($this->jsonVal($json, array('message', 'chat', 'title'), ''))), 0, 40),
          'join_date' => $date,
          'added_by' => $fromId,
          'active' => 1,
        );
        if ($info['telegram_chat_id'] == 0 || $info['chat_name'] == '' || $info['join_date'] <= 0 || $info['added_by'] <= 0) break; // invalid information to add bot to a group -> ignore this message
        $row = $this->db->sql_fetchrow($this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_links WHERE telegram_chat_id = '.$info['telegram_chat_id']));
        if ($row) $this->db->sql_query('UPDATE '.$this->table_prefix."telegram_links ".$this->sqlUpdateStr($info, array('chat_name', 'join_date', 'added_by', 'active'))." WHERE telegram_chat_id = ".$info['telegram_chat_id']);
        else $this->db->sql_query('INSERT INTO '.$this->table_prefix."telegram_links ".$this->sqlInsertStr($info));
        break;
      }
    }

    if ($this->jsonVal($json, array('message', 'left_chat_member', 'id'), 0) == $botId) { // check if this bot is removed from a channel
      if ($telegram_chat_id != 0)
        $this->db->sql_query('UPDATE '.$this->table_prefix."telegram_links SET active = 0 WHERE telegram_chat_id = ".$telegram_chat_id);
    }

    if ($this->jsonVal($json, array('message', 'migrate_to_chat_id'), 0) != 0) { // update to supergroup or whatever...
      $this->db->sql_query('UPDATE '.$this->table_prefix.'telegram_links SET telegram_chat_id = '.intval($this->jsonVal($json, array('message', 'migrate_to_chat_id'), 0)).' * FROM WHERE telegram_chat_id = '.$telegram_chat_id);
    }

    if ($this->jsonVal($json, array('message', 'text'), '', false) != '') { // new incomming or modified message
      $text = '';
      $reply = $this->jsonVal($json, array('message', 'reply_to_message'), null);
      if ($reply && is_array($reply)) {
        $referencePostId = 0;
        $referenceTelegramPostId = $this->jsonVal($reply, array('message_id'), 0);
        $referenceTelegramChatId = $this->jsonVal($reply, array('chat', 'id'), 0);
        if ($referenceTelegramPostId != 0 && $referenceTelegramChatId != 0) {
          $row = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id FROM '.$this->table_prefix.'telegram_messages WHERE telegram_chat_id = '.$referenceTelegramChatId.' AND telegram_message_id = '.$referenceTelegramPostId));
          if ($row)
            $referencePostId = intval($row['post_id']);
        }
        $text =
          '<blockquote'.(($referencePostId == 0)?'':' source="'.$referencePostId.'"').'>'.
            '<cite style="color:gray; margin-left:10px; ">'.htmlspecialchars($this->jsonVal($reply, array('from', 'first_name'), '?')).':</cite>'.
            '<div style="margin-left:10px; ">'.$this->teleramEntitiesDecode($this->jsonVal($reply, array('text'), '', false), $this->jsonVal($reply, array('entities'), null)).'</div>'.
          '</blockquote>';
      }
      $text .= $this->teleramEntitiesDecode($this->jsonVal($json, array('message', 'text'), '', false), $this->jsonVal($json, array('message', 'entities'), null));
      $newPostInfo = $this->newPostFromTelegram(
        $telegram_chat_id,
        $msgId,
        $telegramUser,
        $telegramUserId,
        $editDate,
        $date,
        $text
      );
      $resultMessage = $newPostInfo[0];
    }

    if ($this->jsonVal($json, array('message', 'sticker'), null)) { // sticker - just import fitting emoji as text
      $newPostInfo = $this->newPostFromTelegram(
        $telegram_chat_id,
        $msgId,
        $telegramUser,
        $telegramUserId,
        $editDate,
        $date,
        $this->teleramEntitiesDecode($this->jsonVal($json, array('message', 'sticker', 'emoji'), '[STICKER]', false))
      );
      $resultMessage = $newPostInfo[0];
    }

    if ($this->jsonVal($json, array('message', 'poll'), null)) { // poll created
      $options = $this->jsonVal($json, array('message', 'poll', 'options'), null);
      $optionsText = '';
      if ($options && is_array($options))
        for ($i = 0; $i < count($options); $i++)
          $optionsText .= "\n".html_entity_decode('&#9744;').' '.$this->jsonVal($options[$i], array('text'), '[OPTION]', false);
      $newPostInfo = $this->newPostFromTelegram(
        $telegram_chat_id,
        $msgId,
        $telegramUser,
        $telegramUserId,
        $editDate,
        $date,
        $this->teleramEntitiesDecode($this->jsonVal($json, array('message', 'poll', 'question'), '[POLL]', false).$optionsText),
        '#telegram poll '.$this->jsonVal($json, array('message', 'poll', 'id'), '[POLL]', false)
      );
      $resultMessage = $newPostInfo[0];
    }

    if ($this->jsonVal($json, array('poll'), null)) { // poll update/finish
      $pollId = $this->jsonVal($json, array('poll', 'id'), '[POLL]', false);
      $pollPostRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT post_id, post_text FROM '.$this->table_prefix."posts WHERE post_subject = '".$this->db->sql_escape(utf8_encode_ucr('#telegram poll '.$pollId))."'"));
      $pollPostId = 0;
      $telUserNode = '';
      if ($pollPostRow) {
        $pollPostId = intval($pollPostRow['post_id']);
        $i = strpos($pollPostRow['post_text'], '</telegramuser>');
        if ($i !== false)
          $telUserNode = substr($pollPostRow['post_text'], 0, $i+15);
      }
      if ($pollPostId != 0) {
        $options = $this->jsonVal($json, array('poll', 'options'), null);
        $totalVotes = $this->jsonVal($json, array('poll', 'total_voter_count'), 0);
        $pollText = utf8_encode_ucr(htmlspecialchars($this->jsonVal($json, array('poll', 'question'), '[POLL]', false))).'<br>';
        if ($options && is_array($options))
          for ($i = 0; $i < count($options); $i++) {
            $votes = $this->jsonVal($options[$i], array('voter_count'), 0);
            $pollText .=
              '<div style="display:inline-block; height:1em; width:100px; margin-right:5px; position:relative; border: 1px solid #333; border-radius:3px; overflow:hidden; ">'.
                '<div style="position:absolute; height:100%; width:'.(($totalVotes > 0)?$votes/$totalVotes*100:0).'%; background-color:#333; "></div>'.
              '</div>'.
              '<div style="display:inline-block; margin-right:5px; width:30px; ">'.$votes.'</div>'.
              '<div style="display:inline-block; width:calc(100% - 150px); ">'.utf8_encode_ucr(htmlspecialchars($this->jsonVal($options[$i], array('text'), '[OPTION]', false))).'</div><br>';
          }
        $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = '".$telUserNode.'<t>'.$this->db->sql_escape($pollText)."</t>' WHERE post_id = ".$pollPostId);
        $resultMessage = 'Poll updated.';
      } else $resultMessage = 'Poll not found!';
    }

    $fotoSizes = $this->jsonVal($json, array('message', 'photo'), null);
    if ($fotoSizes) { // single image with caption
      // find maximum foto
      $maxSize = -1;
      $maxId = '';
      for ($i = 0; $i < count($fotoSizes); $i++) {
        $si = $this->jsonVal($fotoSizes[$i], array('file_size'), 0);
        if ($si > $maxSize) {
          $maxSize = $si;
          $maxId = $this->jsonVal($fotoSizes[$i], array('file_id'), $maxId);
        }
      }
      if ($maxId == '')
        $importRes = ['NO_MAX_SIZE_PHOTO_ID'];
      else
        $importRes = $this->telegramFileImport(
          $telegram_chat_id,
          $msgId,
          $telegramUser,
          $telegramUserId,
          $editDate,
          $date,
          $this->jsonVal($json, array('message', 'media_group_id'), ''),
          $maxId,
          $this->teleramEntitiesDecode($this->jsonVal($json, array('message', 'caption'), '', false), $this->jsonVal($json, array('message', 'caption_entities'), null)),
          ''
        );
      if (count($importRes) == 0 || $importRes[0] != 'SUCCESS') $resultMessage = join(', ', $importRes);
    }

    if ($this->jsonVal($json, array('message', 'document'), null)) { // document appendix, eventually an animation
      $fileId = $this->jsonVal($json, array('message', 'document', 'file_id'), '');
      if ($fileId == '')
        $importRes = ['NO_FILE_DOWNLOAD_ID'];
      else
        $importRes = $this->telegramFileImport(
          $telegram_chat_id,
          $msgId,
          $telegramUser,
          $telegramUserId,
          $editDate,
          $date,
          $this->jsonVal($json, array('message', 'media_group_id'), ''),
          $fileId,
          '',
          $this->jsonVal($json, array('message', 'document', 'file_name'), '')
        );
      if (count($importRes) == 0 || $importRes[0] != 'SUCCESS')
        $resultMessage = join(', ', $importRes);
      else if ($this->jsonVal($json, array('message', 'animation', 'mime_type'), '') == 'video/mp4') { // add animation preview
        $video = // this will show the video in post and hide the attachment box for this file
          '<video loop muted autoplay style="max-width:100%;"><source src="/download/file.php?id='.intval($importRes[1]).'" type="video/mp4"></video>'.
          $this->attachmentHideScript('video', $importRes[1]);
        $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = CONCAT(post_text, '".$this->db->sql_escape($video)."') WHERE post_id = ".intval($importRes[2]));
      }
    }

    if ($this->jsonVal($json, array('message', 'voice'), null)) { // an audio file
      $fileId = $this->jsonVal($json, array('message', 'voice', 'file_id'), '');
      if ($fileId == '')
        $importRes = ['NO_FILE_DOWNLOAD_ID'];
      else
        $importRes = $this->telegramFileImport(
          $telegram_chat_id,
          $msgId,
          $telegramUser,
          $telegramUserId,
          $editDate,
          $date,
          $this->jsonVal($json, array('message', 'media_group_id'), ''),
          $fileId
        );
      if (count($importRes) == 0 || $importRes[0] != 'SUCCESS')
        $resultMessage = join(', ', $importRes);
      else if ($this->jsonVal($json, array('message', 'voice', 'mime_type'), '') == 'audio/ogg') { // add animation preview
        $audio = // this will show the audio in post and hide the attachment box for this file
          '<audio controls style="width:300px;"><source src="/download/file.php?id='.intval($importRes[1]).'" type="audio/ogg"></audio>'.
          $this->attachmentHideScript('audio', $importRes[1]);
        $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = CONCAT(post_text, '".$this->db->sql_escape($audio)."') WHERE post_id = ".intval($importRes[2]));
      }
    }

    if ($this->jsonVal($json, array('message', 'video'), null)) { // video file
      $fileId = $this->jsonVal($json, array('message', 'video', 'file_id'), '');
      if ($fileId == '')
        $importRes = ['NO_FILE_DOWNLOAD_ID'];
      else
        $importRes = $this->telegramFileImport(
          $telegram_chat_id,
          $msgId,
          $telegramUser,
          $telegramUserId,
          $editDate,
          $date,
          $this->jsonVal($json, array('message', 'media_group_id'), ''),
          $fileId,
          $this->teleramEntitiesDecode($this->jsonVal($json, array('message', 'caption'), '', false), $this->jsonVal($json, array('message', 'caption_entities'), null)),
          $this->jsonVal($json, array('message', 'video', 'file_name'), '')
        );
      if (count($importRes) == 0 || $importRes[0] != 'SUCCESS')
        $resultMessage = join(', ', $importRes);
      else if ($this->jsonVal($json, array('message', 'video', 'mime_type'), '') == 'video/mp4') { // add animation preview
        $video = // this will show the video in post and hide the attachment box for this file
          '<video loop muted autoplay style="max-width:100%;"><source src="/download/file.php?id='.intval($importRes[1]).'" type="video/mp4"></video>'.
          $this->attachmentHideScript('video', $importRes[1]);
        $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_text = CONCAT(post_text, '".$this->db->sql_escape($video)."') WHERE post_id = ".intval($importRes[2]));
      }
    }

    return new \Symfony\Component\HttpFoundation\Response($resultMessage, 200);
  }

  private function telegramOp ($opQuery) {
    $result = null;
    $context = stream_context_create(array('http' => array('ignore_errors' => true)));
    $response = @file_get_contents('https://api.telegram.org/bot'.$this->config['telegram_bot_key'].'/'.$opQuery, false, $context);
    if ($response) {
      try {
        $result = json_decode($response, true);
      } catch (Exception $e) { $result = null; }
    }
    return $result;
  }

  private function treeContainsOnly ($tree, $allowed) {
    for ($i = 0; $i < count($tree); $i++)
      if (array_search($tree[$i], $allowed) === false)
        return false;
    return true;
  }

  public function telegramDownloadForward ($fileId) {
    $url = $this->telegramDownloadLink($fileId);
    if ($url[0] == '') return new \Symfony\Component\HttpFoundation\Response('Download Error: '.$url[1], 500);
    return new \Symfony\Component\HttpFoundation\Response('Forwarding to telegram download link...', 303, ['location' => $url[0]]);
  }

  public function sendChannels ($inputBridge, $inputInfo, $postMode, $template, $message, $forumId, $topicId, $postId) {
    if (getType($forumId) != 'integer' || getType($topicId) != 'integer' || getType($postId) != 'integer') throw 'Type Missmatch';
    $telMsgs = null;
    $sqlResponse = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_links WHERE forum_link_id = '.$topicId.' AND (link_type = 2 OR link_type = 3) AND active = 1');
    $err = array();
    // init variables
    $referenceTelegramPostId = 0;
    $referenceTelegramChatId = 0;
    $quoteVar = '';
    $text = '';
    $textRmQuote = '';
    if ($postMode != 'delete') {
      // compose text
      $template->assign_vars(array(
        'msg' => $message,
        'e2e_availlable' => false,
        'will_be_e2e' => false,
        'formatting_level' => 2,
        'bridge' => self::SHORT_NAME,
      ));
      $text = trim($template->assign_display('html_chat_forwarding'));
      $dom = new \DOMDocument();
      @$dom->loadHTML('<?xml encoding="utf-8" ?>'.$text);
      $text = $this->formatHTMLForTelegram($dom->documentElement, array());
      $dom = null;
      // search for reply-quote and reference post
      $i1 = strpos($message, '<blockquote source="');
      if ($i1 !== false) {
        $i2 = strpos($message, '"', $i1+20);
        $i3 = 0;
        $referencePostId = 0;
        if ($i2 !== false)
          $referencePostId = intval(substr($message, $i1+20, $i2-$i1-20));
        if ($referencePostId != 0) {
          $i3 = strpos($message, '</blockquote>', $i2);
            $refRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT telegram_chat_id, telegram_message_id FROM '.$this->table_prefix.'telegram_messages WHERE post_id = '.$referencePostId));
          if ($refRow && $i3 !== false) {
            $referenceTelegramChatId = intval($refRow['telegram_chat_id']);
            $referenceTelegramPostId = intval($refRow['telegram_message_id']);
          }
        }
        if ($referenceTelegramPostId != 0) {
          $dom->loadHTML('<?xml encoding="utf-8" ?>'.substr($message, 0, $i1).substr($message, $i3+13));
          $template->assign_vars(array('msg' => $this->formatHTMLForTelegram($dom->documentElement, array())));
          $textRmQuote = trim($template->assign_display('html_chat_forwarding'));
          $quoteVar = '&reply_to_message_id='.$referenceTelegramPostId;
        }
      }
    }
    // post to all channels
    while ($row = $this->db->sql_fetchrow($sqlResponse)) {
      if ($inputBridge == self::SHORT_NAME && intval($row['telegram_chat_id']) == intval($inputInfo)) continue; // don't re-post a messige from telegram to origin channel
      if ($postMode == 'edit' || $postMode == 'delete') {
        if (!$telMsgs) $telMsgs = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_messages WHERE post_id = '.intval($postId));
        foreach ($telMsgs as $mRow)
          if ($mRow['telegram_chat_id'] == $row['telegram_chat_id']) { // to have this in the double loop ensures, that no telegram messages of deleted links are updated
            if ($postMode == 'delete') {
              $this->telegramOp('deleteMessage?chat_id='.intval($row['telegram_chat_id']).'&message_id='.intval($mRow['telegram_message_id']));
            } else { // edit
              $json = $this->telegramOp('editMessageText?chat_id='.intval($row['telegram_chat_id']).'&message_id='.intval($mRow['telegram_message_id']).'&text='.rawurlencode($text).'&parse_mode=HTML');
              if (!$json || !$json['ok'])
                $json = $this->telegramOp('sendMessage?chat_id='.intval($row['telegram_chat_id']).'&reply_to_message_id='.intval($mRow['telegram_message_id']).'&text='.rawurlencode(trim($template->assign_display('html_edit_post_info'))).'&parse_mode=HTML');
            }
            break; // there can only be one corresponding telegram message per link
          }
      } else {
        $json = $this->telegramOp('sendMessage?chat_id='.intval($row['telegram_chat_id']).'&text='.rawurlencode(($referenceTelegramChatId == intval($row['telegram_chat_id']))?$textRmQuote:$text).'&parse_mode=HTML'.$quoteVar);
        $msgId = intval($this->jsonVal($json, array('result', 'message_id'), 0));
        if ($msgId == 0) $err[] = $this->jsonVal($json, array('description'), '');
        else $this->db->sql_query('INSERT INTO '.$this->table_prefix.'telegram_messages (telegram_chat_id, telegram_message_id, post_id) VALUES ('.intval($row['telegram_chat_id']).', '.$msgId.', '.intval($postId).')');
      }
    }
    return $err;
  }

  public function deletePost ($inputBridge, $inputInfo, $forumId, $topicId, $postId, $reason) {
    return $this->sendChannels($inputBridge, $inputInfo, 'delete', null, '', $forumId, $topicId, $postId);
  }

  public function listAvaillableChannels () {
    $str = '';
    $res = $this->db->sql_query('SELECT * FROM '.$this->table_prefix.'telegram_links WHERE forum_link_id = 0 AND active = 1');
    while ($row = $this->db->sql_fetchrow($res))
      $str .= $row['telegram_chat_id']."\t".rawurlencode($row['chat_name'])."\t".$row['join_date']."\t".$row['added_by']."\n";
    return new \Symfony\Component\HttpFoundation\Response($str, 200);
  }


}
